"""Base functionality for optical analyses.

This module provides the `analysis` decorator for optical analyses.
"""

from __future__ import annotations

import inspect
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, TypeVar, cast

from visisipy.backend import BaseBackend, get_backend

if TYPE_CHECKING:
    from visisipy.models import EyeModel

__all__ = ("_AUTOMATIC_BACKEND", "analysis")


_AUTOMATIC_BACKEND = cast(type[BaseBackend], object())
"""Default sentinel value for the backend parameter in analysis functions.

The `backend` parameter in analysis functions is not optional, but does not need to be specified if the default
backend is used. However, not specifying a default argument causes IDEs to warn about missing arguments. To avoid this,
a sentinel value is used to indicate that the default backend should be used. Although this value keeps the type
checker happy, it is not a valid backend value and is only intended for internal use.
"""


def _validate_analysis_signature(function: Callable[..., Any | tuple[Any, Any]]) -> None:
    """Validate the signature of an analysis function.

    The first parameter must be 'model' with type 'EyeModel | None'. The last parameter must be 'return_raw_result'
    with type 'bool'.

    Parameters
    ----------
    function : Callable
        The analysis function to validate.

    Raises
    ------
    ValueError
        If the signature of the analysis function is invalid.
    """

    signature = inspect.signature(function)
    parameter_names = list(signature.parameters)

    if parameter_names[0] != "model":
        raise ValueError("The first parameter of an analysis function must be 'model'.")

    if signature.parameters["model"].annotation != "EyeModel | None":
        raise ValueError(
            f"The 'model' parameter of an analysis function must have type 'EyeModel | None', "
            f"got '{signature.parameters['model'].annotation}'"
        )

    if "backend" not in parameter_names:
        raise ValueError(
            "The analysis function must have a keyword-only 'backend' parameter of type 'type[BaseBackend]'."
        )

    if signature.parameters["backend"].kind.name != "KEYWORD_ONLY":
        raise ValueError("The 'backend' parameter of an analysis function must be keyword-only.")

    if signature.parameters["backend"].annotation != "type[BaseBackend]":
        raise ValueError(
            f"The 'backend' parameter of an analysis function must have type 'type[BaseBackend]', "
            f"got '{signature.parameters['backend'].annotation}'"
        )

    if "return_raw_result" not in parameter_names:
        raise ValueError("The analysis function must have a keyword-only 'return_raw_result' parameter of type 'bool'.")

    if signature.parameters["return_raw_result"].kind.name != "KEYWORD_ONLY":
        raise ValueError("The 'return_raw_result' parameter of an analysis function must be keyword-only.")

    if signature.parameters["return_raw_result"].annotation != "bool":
        raise ValueError(
            f"The 'return_raw_result' parameter of an analysis function must have type 'bool', "
            f"got '{signature.parameters['return_raw_result'].annotation}'"
        )


def _build_model(model: EyeModel, backend: type[BaseBackend]) -> None:
    """Build the model in OpticStudio if it is not already built.

    Parameters
    ----------
    model : Any
        The model to build.
    """
    if backend.model is None or backend.model.eye_model is not model:
        backend.build_model(model)


T1 = TypeVar("T1")
T2 = TypeVar("T2")


def analysis(function: Callable[..., T1 | tuple[T1, T2]]) -> Callable:
    """Decorator for analysis functions.

    This decorator is used to mark a function as an analysis function. Analysis functions are used to perform various
    analyses on the optical system. This decorator passes the model and backend to the function and ensures the
    model is built. Furthermore, it validates the function signature to ensure it has the correct parameters.

    Every analysis must have at least the following parameters:

    - `model: EyeModel | None` must be the first parameter. If not specified by the user, the eye model that is currently
        built in the backend will be used.
    - `return_raw_result: bool` must be a keyword-only parameter. If `True`, the function will return a tuple with the
        result and the raw result returned by the backend.
    - `backend: type[BaseBackend]` must be a keyword-only parameter. If not specified by the user, the currently configured
        backend will be used.

    Parameters
    ----------
    function : Callable
        The analysis function.

    Returns
    -------
    Callable
        The analysis function.

    Raises
    ------
    ValueError
        If the signature of the analysis function is invalid.
    """
    _validate_analysis_signature(function)

    @wraps(function)
    def analysis_wrapper(
        model: EyeModel | None = None,
        *args: Any,
        return_raw_result: bool = False,
        backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
        **kwargs: Any,
    ) -> T1 | tuple[T1, T2]:
        if backend is None or backend is _AUTOMATIC_BACKEND:
            backend = get_backend()

        if model is not None:
            _build_model(model, backend)
        elif backend.model is None:
            message = (
                "No model was provided and no model is currently built in the backend."
                " Please provide a model or first build one using `EyeModel.build`."
            )
            raise ValueError(message)

        result, raw_result = function(model, *args, return_raw_result=return_raw_result, backend=backend, **kwargs)

        if return_raw_result:
            return result, raw_result

        return result

    return analysis_wrapper



"""Analyze optical eye models in OpticStudio."""

from __future__ import annotations

from visisipy.opticstudio.backend import OpticStudioBackend, OpticStudioSettings
from visisipy.opticstudio.models import OpticStudioEye

__all__ = ("OpticStudioBackend", "OpticStudioEye", "OpticStudioSettings")



from __future__ import annotations

import math

import pytest

from tests.helpers import build_args
from visisipy.models import EyeModel
from visisipy.types import SampleSize

pytestmark = [pytest.mark.needs_opticstudio]


class MockPupil:
    def __init__(self):
        self.semi_diameter = 1.0
        self.changed_semi_diameter = False

    @property
    def semi_diameter(self):
        return self._semi_diameter

    @semi_diameter.setter
    def semi_diameter(self, value):
        self.changed_semi_diameter = True
        self._semi_diameter = value


class MockOpticstudioModel:
    def __init__(self):
        self._pupil = MockPupil()

    @property
    def pupil(self):
        return self._pupil


class TestRefractionAnalysis:
    @pytest.mark.parametrize(
        "field_coordinate,wavelength,sampling,pupil_diameter,field_type,use_higher_order_aberrations",
        [
            (None, None, None, None, None, None),
            ((0, 0), 0.543, None, None, None, None),
            ((1, 1), 0.632, 64, None, None, None),
            ((math.pi, math.tau), 0.543, "64x64", 0.5, None, None),
            ((0, 0), 0.543, "64x64", 0.5, "angle", None),
            ((1, 1), 0.632, SampleSize(32), 0.5, "object_height", None),
            ((0, 0), 0.543, "32x32", 0.5, "angle", True),
            ((1, 1), 0.632, "64x64", 0.5, "object_height", False),
        ],
    )
    def test_refraction(
        self,
        field_coordinate,
        wavelength,
        sampling,
        pupil_diameter,
        field_type,
        use_higher_order_aberrations,
        opticstudio_analysis,
        opticstudio_backend,
        monkeypatch,
    ):
        opticstudio_backend.build_model(
            EyeModel(), object_distance=10 if field_type == "object_height" else float("inf")
        )
        # monkeypatch.setattr(opticstudio_analysis.backend, "model", MockOpticstudioModel())

        args = build_args(
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            sampling=sampling,
            pupil_diameter=pupil_diameter,
            field_type=field_type,
            use_higher_order_aberrations=use_higher_order_aberrations,
            non_null_defaults={"sampling", "field_type", "use_higher_order_aberrations"},
        )

        assert opticstudio_analysis.refraction(**args)

    @pytest.mark.parametrize(
        "pupil_diameter,changed_pupil_diameter",
        [
            (None, False),
            (0.5, True),
        ],
    )
    def test_refraction_change_pupil(self, opticstudio_analysis, pupil_diameter, changed_pupil_diameter, monkeypatch):
        monkeypatch.setattr(opticstudio_analysis.backend, "model", MockOpticstudioModel())

        assert not opticstudio_analysis.backend.model.pupil.changed_semi_diameter

        opticstudio_analysis.refraction(pupil_diameter=pupil_diameter)

        assert opticstudio_analysis.backend.model.pupil.changed_semi_diameter == changed_pupil_diameter
        assert opticstudio_analysis.backend.model.pupil.semi_diameter == 1.0



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from visisipy.optiland.backend import OptilandBackend
    from visisipy.types import FieldType

__all__ = ("set_field", "set_wavelength")


def set_wavelength(
    backend: type[OptilandBackend],
    wavelength: float | None = None,
) -> float:
    """Set the wavelength in the Optiland backend.

    Parameters
    ----------
    backend : OptilandBackend
        Reference to the Optiland backend.
    wavelength : float | None, optional
        The wavelength to set. If `None`, the first wavelength in the system will be used. Defaults to `None`.

    Returns
    -------
    float
        The wavelength that was set. If `None` was passed, the first wavelength in the system is returned.
    """
    if wavelength is None:
        wavelength = backend.get_wavelengths()[0]
    else:
        backend.set_wavelengths([wavelength])

    return wavelength


def set_field(
    backend: type[OptilandBackend],
    field_coordinate: tuple[float, float] | None = None,
    field_type: FieldType = "angle",
) -> tuple[float, float]:
    """Set the field coordinate in the Optiland backend.

    Parameters
    ----------
    backend : OptilandBackend
        Reference to the Optiland backend.
    field_coordinate : tuple[float, float] | None, optional
        The field coordinate to set. If `None`, the first field in the system will be used. Defaults to `None`.
    field_type : str, optional
        The type of field to be used when setting the field coordinate. Defaults to "angle".

    Returns
    -------
    tuple[float, float]
        The normalized field coordinate of the field that was set.
    """
    if field_coordinate is not None:
        backend.set_fields([field_coordinate], field_type=field_type)

    return backend.get_optic().fields.get_field_coords()[0]



from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple

import numpy as np

if TYPE_CHECKING:
    import pandas as pd


def get_ray_output_angle(
    df: pd.DataFrame,
    reference_point: tuple[float, float] = (0, 0),
    coordinate="y",
):
    """Calculate the output angle of a ray with respect to the optical axis and a reference point."""
    x0, y0 = reference_point
    x1, y1 = df.loc[len(df) - 1, "z"], df.loc[len(df) - 1, coordinate]

    return np.rad2deg(np.arctan2(y1 - y0, x1 - x0))


class InputOutputAngles(NamedTuple):
    input_angle_field: float
    output_angle_pupil: float
    output_angle_np2: float
    output_angle_retina_center: float = None
    output_angle_navarro_np2: float = None
    location_np2: float = None
    location_retina_center: float = None
    patient: int | str | None = None

    @classmethod
    def from_ray_trace_result(
        cls,
        raytrace_result: pd.DataFrame,
        np2: float,
        np2_navarro: float | None = None,
        retina_center: float | None = None,
        patient: int | None = None,
        coordinate="y",
    ) -> InputOutputAngles:
        return cls(
            input_angle_field=raytrace_result.field[0][1],
            output_angle_pupil=get_ray_output_angle(raytrace_result, reference_point=(0, 0), coordinate=coordinate),
            output_angle_np2=get_ray_output_angle(raytrace_result, reference_point=(np2, 0), coordinate=coordinate),
            output_angle_retina_center=(
                get_ray_output_angle(
                    raytrace_result,
                    reference_point=(retina_center, 0),
                    coordinate=coordinate,
                )
                if retina_center is not None
                else None
            ),
            output_angle_navarro_np2=(
                get_ray_output_angle(
                    raytrace_result,
                    reference_point=(np2_navarro, 0),
                    coordinate=coordinate,
                )
                if np2_navarro is not None
                else None
            ),
            location_np2=np2,
            location_retina_center=retina_center or None,
            patient=patient,
        )



import logging
import os
from datetime import datetime
from subprocess import run

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Visisipy"
year = datetime.now().year  # noqa: DTZ005
copyright_year = str(year) if year == 2025 else f"2025 - {year}"  # noqa: PLR2004
copyright = f"{copyright_year}, Corné Haasjes, Luc van Vught, Jan-Willem Beenakker"  # noqa: A001
author = "Corné Haasjes, Luc van Vught, Jan-Willem Beenakker"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["autoapi.extension", "myst_nb", "sphinx_design"]
myst_enable_extensions = ["colon_fence"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]

html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/MREYE-LUMC/visisipy",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/visisipy",
            "icon": "fa-brands fa-python",
        },
        {
            "name": "MReye.nl",
            "url": "https://mreye.nl",
            "icon": "https://mreye.nl/icon.png",
            "type": "url",
        },
    ],
    "logo": {
        "text": "Visisipy",
    },
    "show_toc_level": 1,
    "use_edit_page_button": True,
    "secondary_sidebar_items": {
        "**": ["page-toc"],
        "user_guide/**": ["page-toc", "download-notebook"],
        "examples/**/**": ["page-toc", "download-notebook"],
    },
}

if os.getenv("READTHEDOCS") == "True":
    git_branch = os.getenv("READTHEDOCS_GIT_IDENTIFIER", "main")
else:
    try:
        git_branch = run(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],  # noqa: S607
            capture_output=True,
            text=True,
            check=True,
        ).stdout.strip()
    except:  # noqa: E722
        git_branch = "main"

logger.info("Building documentation for branch %s", git_branch)

html_context = {
    "edit_page_url_template": "{{ github_url }}/{{ github_user }}/{{ github_repo }}/tree/{{ github_version }}/{{ doc_path }}{{ file_name }}",
    "edit_page_provider_name": "GitHub",
    "github_user": "MREYE-LUMC",
    "github_repo": "visisipy",
    "github_version": git_branch,
    "doc_path": "docs",
}

# -- Options for myst-nb ----------------------------------------------------
# https://myst-nb.readthedocs.io/en/latest/configuration.html

if os.getenv("RUN_NOTEBOOKS") == "false":
    nb_execution_mode = "off"
    logger.info("Skipping notebook execution")
elif os.getenv("READTHEDOCS") == "True":
    nb_execution_mode = "force"
    logger.info("Using notebook execution mode '%s'", nb_execution_mode)
else:
    nb_execution_mode = "cache"
    logger.info("Using notebook execution mode '%s'", nb_execution_mode)


if os.getenv("READTHEDOCS"):
    # Do not build notebooks that depend on OpticStudio
    execution_excludepatterns = [
        "*examples/Patient-specific mapping of fundus photographs to three-dimensional ocular imaging/*",
        "*examples/Backend comparison/*",
    ]

# -- Options for autoapi ----------------------------------------------------
# https://sphinx-autoapi.readthedocs.io/en/latest/

autoapi_dirs = ["../visisipy"]
autoapi_root = "api"
autoapi_options = [
    "members",
    "show-inheritance",
    "show-module-summary",
]



from __future__ import annotations

import pytest

from tests.helpers import build_args
from visisipy import EyeModel


class TestFFTPSFAnalysis:
    @pytest.mark.parametrize(
        "field_coordinate,wavelength,field_type,sampling",
        [
            (None, None, None, None),
            (None, None, "object_height", 64),
            ((0, 0), 0.550, "angle", 128),
            ((10, 5), 0.400, "object_height", 256),
            ((5.5, 5.5), 0.550, "angle", 512),
        ],
    )
    def test_fft_psf(
        self,
        optiland_backend,
        optiland_analysis,
        field_coordinate,
        wavelength,
        field_type,
        sampling,
    ):
        optiland_backend.build_model(EyeModel(), object_distance=10 if field_type == "object_height" else float("inf"))

        args = build_args(
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            field_type=field_type,
            sampling=sampling,
            non_null_defaults={"field_type", "sampling"},
        )

        assert optiland_analysis.fft_psf(**args)


class TestHuygensPSFAnalysis:
    @pytest.mark.parametrize(
        "field_coordinate,wavelength,field_type,pupil_sampling,image_sampling",
        [
            (None, None, None, None, None),
            (None, None, "object_height", 32, 32),
            ((0, 0), 0.550, "angle", 64, 128),
            ((10, 5), 0.400, "object_height", 256, 128),
            ((5.5, 5.5), 0.550, "angle", 64, 64),
        ],
    )
    def test_fft_psf(
        self,
        optiland_backend,
        optiland_analysis,
        field_coordinate,
        wavelength,
        field_type,
        pupil_sampling,
        image_sampling,
    ):
        optiland_backend.build_model(EyeModel(), object_distance=10 if field_type == "object_height" else float("inf"))

        args = build_args(
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            field_type=field_type,
            pupil_sampling=pupil_sampling,
            image_sampling=image_sampling,
            non_null_defaults={"field_type", "pupil_sampling", "image_sampling"},
        )

        assert optiland_analysis.huygens_psf(**args)


class TestStrehlRatioAnalysis:
    @pytest.mark.parametrize(
        "field_coordinate,wavelength,field_type,sampling",
        [
            (None, None, None, None),
            (None, None, "object_height", 64),
            ((0, 0), 0.550, "angle", 128),
            ((10, 5), 0.400, "object_height", 256),
            ((5.5, 5.5), 0.550, "angle", 64),
        ],
    )
    @pytest.mark.parametrize(
        "psf_type",
        ["fft", "huygens"],
    )
    def test_strehl_ratio(
        self,
        optiland_backend,
        optiland_analysis,
        field_coordinate,
        wavelength,
        field_type,
        sampling,
        psf_type,
    ):
        optiland_backend.build_model(EyeModel(), object_distance=10 if field_type == "object_height" else float("inf"))

        args = build_args(
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            field_type=field_type,
            sampling=sampling,
            psf_type=psf_type,
            non_null_defaults={"field_type", "sampling"},
        )

        assert optiland_analysis.strehl_ratio(**args)

    def test_invalid_psf_type(self, optiland_backend, optiland_analysis):
        optiland_backend.build_model(EyeModel(), object_distance=float("inf"))

        with pytest.raises(NotImplementedError, match="PSF type 'invalid' is not implemented"):
            optiland_analysis.strehl_ratio(psf_type="invalid")



"""Optical surfaces for Optiland."""

from __future__ import annotations

import weakref
from functools import reduce, singledispatch
from operator import attrgetter
from typing import TYPE_CHECKING, Generic, TypeVar, Union

import optiland.materials
from optiland.materials import AbbeMaterial, IdealMaterial, Material

from visisipy.models import BaseSurface
from visisipy.models import NoSurface as OptilandNoSurface
from visisipy.models.geometry import (
    NoSurface,
    StandardSurface,
    Stop,
    Surface,
    ZernikeStandardPhaseSurface,
    ZernikeStandardSagSurface,
)
from visisipy.models.materials import MaterialModel

if TYPE_CHECKING:
    import optiland.surfaces
    from optiland.optic import Optic


PropertyType = TypeVar("PropertyType")


class OptilandSurfaceProperty(Generic[PropertyType]):
    """Descriptor for Optiland surface properties.

    This descriptor is used to access and set properties of the Optiland surface.
    """

    def __init__(self, name: str):
        self.name = name
        self.attrgetter = attrgetter(name)

    def __get__(self, obj: OptilandSurface, objtype=None) -> PropertyType:
        if obj.surface is None:
            return None

        return self.attrgetter(obj.surface)

    def __set__(self, obj: OptilandSurface, value: PropertyType) -> None:
        if obj.surface is None:
            return

        name_split = self.name.split(".")

        if len(name_split) == 1:
            setattr(obj.surface, self.name, value)
        else:
            setattr(
                reduce(getattr, name_split[:-1]),
                name_split[-1],
                value,
            )


class _built_only_property(property):  # noqa: N801
    """Property that can only be accessed after the surface has been built."""

    def __get__(self, obj: OptilandSurface, objtype=None):
        if not obj._is_built:  # noqa: SLF001
            return None

        return super().__get__(obj, objtype)

    def __set__(self, obj: OptilandSurface, value) -> None:
        if not obj._is_built:  # noqa: SLF001
            message = "Cannot set attribute of non-built surface."
            raise AttributeError(message)

        super().__set__(obj, value)


class OptilandSurface(BaseSurface):
    """Sequential surface in Optiland.

    Note that, unlike OpticStudio surfaces, it is not possible to update surfaces after they have been built.
    """

    _TYPE: str = "Standard"

    def __init__(
        self,
        comment: str,
        *,
        radius: float = float("inf"),
        thickness: float = 0.0,
        semi_diameter: float | None = None,
        conic: float = 0.0,
        material: MaterialModel | str | None = None,
        is_stop: bool | None = None,
    ):
        """Create a new Optiland surface.

        Parameters
        ----------
        comment : str
            Comment for the surface.
        radius : float
            Radius of curvature of the surface, in mm. Defaults to infinity (flat surface).
        thickness : float
            Thickness of the surface, in mm. Defaults to 0.0 mm.
        semi_diameter : float, optional
            Semi-diameter of the surface aperture, in mm. Defaults to `None`, which means the semi-diameter will be
            determined by Optiland.
        conic : float
            Conic constant of the surface. Defaults to 0.0 (spherical surface).
        material : MaterialModel | str, optional
            Material of the surface. This can be either a string representing the name of the material or a
            MaterialModel instance. Defaults to `None`, which means the material is assumed to be air.
        is_stop : bool, optional
            If `True`, the surface is treated as a stop.
        """
        self._comment = comment
        self._radius = radius
        self._thickness = thickness
        self._semi_diameter = semi_diameter
        self._conic = conic
        self._material = material
        self._is_stop = is_stop

        self._surface: optiland.surfaces.Surface | None = None
        self._optic: Optic | None = None
        self._index: int | None = None
        self._is_built: bool = False

    @property
    def surface(self) -> optiland.surfaces.Surface | None:
        """Optiland surface object.

        This property only has a value if the surface has been built.
        """
        return self._surface

    @_built_only_property
    def comment(self) -> str:
        """Comment for the surface."""
        return self.surface.comment  # type: ignore

    @comment.setter
    def comment(self, value: str) -> None:
        """Set the comment for the surface."""
        self.surface.comment = value  # type: ignore

    @_built_only_property
    def radius(self) -> float:
        """Radius of the surface."""
        return self._optic.surface_group.radii[self._index]  # type: ignore

    @_built_only_property
    def thickness(self) -> float:
        """Thickness of the surface."""
        if self._index == self._optic.surface_group.num_surfaces - 1:
            # Last surface in the system, return 0.0
            return 0.0

        return self._optic.surface_group.get_thickness(self._index)[0]  # type: ignore

    @_built_only_property
    def semi_diameter(self) -> float | None:
        """Semi-diameter of the surface."""
        return self.surface.semi_aperture  # type: ignore

    @_built_only_property
    def conic(self) -> float:
        """Conic constant of the surface."""
        return self._optic.surface_group.conic[self._index]  # type: ignore

    @_built_only_property
    def material(self) -> MaterialModel | str | None:
        """Material of the surface."""
        return self._get_material()

    @_built_only_property
    def is_stop(self) -> bool:
        """Flag indicating if the surface is a stop."""
        return self.surface.is_stop  # type: ignore

    def _get_material(self) -> MaterialModel | str | None:
        """Get the material of the surface."""
        if not self._is_built:
            return None

        if isinstance(self.surface.material_post, IdealMaterial):  # type: ignore
            return MaterialModel(
                refractive_index=self.surface.material_post.index,  # type: ignore
                abbe_number=0.0,
                partial_dispersion=0.0,
            )

        if isinstance(self.surface.material_post, AbbeMaterial):  # type: ignore
            return MaterialModel(
                refractive_index=self.surface.material_post.index,  # type: ignore
                abbe_number=self.surface.material_post.abbe,  # type: ignore
                partial_dispersion=0.0,
            )

        if isinstance(self.surface.material_post, Material):  # type: ignore
            return self.surface.material_post.name  # type: ignore

        raise TypeError(
            f"Unsupported material type: {type(self.surface.material_post)}"  # type: ignore
        )

    @staticmethod
    def _convert_material(
        material: MaterialModel | str | None,
    ) -> str | IdealMaterial | AbbeMaterial:
        if isinstance(material, str):
            return material

        if material is None:
            return "Air"

        if isinstance(material, MaterialModel):
            if material.abbe_number == 0:
                return IdealMaterial(n=material.refractive_index, k=0.0)

            return AbbeMaterial(n=material.refractive_index, abbe=material.abbe_number)

        raise TypeError("'material' must be MaterialModel or str.")

    def build(self, optic: Optic, *, position: int, replace_existing: bool = False) -> int:
        """Create the surface in Optiland.

        Create the surface in the provided `Optic` object at the specified `position`.
        By default, a new surface will be created. If `replace_existing` is `True`, the existing surface at the
        specified position will be replaced.

        Parameters
        ----------
        optic : Optic
            The Optic object to which the surface will be added.
        position : int
            The index at which the surface will be added, starting at 0 for the object surface.
        replace_existing : bool
            If `True`, replace an existing surface instead of inserting a new one. Defaults to `False`.

        Returns
        -------
        int
            The index of the created surface. Subsequent surfaces should be after this index.
        """
        optic.add_surface(
            index=position,
            radius=self._radius,
            thickness=self._thickness,
            conic=self._conic,
            material=self._convert_material(self._material),
            is_stop=bool(self._is_stop),
            comment=self._comment,
        )

        if replace_existing:
            optic.surface_group.surfaces.pop(position + 1)

        self._surface = optic.surface_group.surfaces[position]

        if self._semi_diameter is not None:
            self.surface.set_semi_aperture(self._semi_diameter)  # type: ignore

        self._optic = weakref.proxy(optic)
        self._index = position
        self._is_built = True

        return position


@singledispatch
def make_surface(surface: Surface, material: str | MaterialModel, comment: str = "") -> OptilandSurface:
    """Create an `OptilandSurface` instance from a given `Surface` instance.

    Parameters
    ----------
    surface : Surface
        The Surface instance from which to create the OptilandSurface instance.
    material : str | MaterialModel
        The material of the surface. This can be either a string representing the name
        of the material or a MaterialModel instance.
    comment : str, optional
        A comment to be associated with the surface. This is an empty string by default.

    Returns
    -------
    OptilandSurface
        The created OptilandSurface instance.
    """
    return OptilandSurface(comment=comment, thickness=surface.thickness, material=material)


@make_surface.register
def _make_surface(
    surface: StandardSurface,
    material: Union[str, MaterialModel],  # noqa: UP007
    comment: str = "",
) -> OptilandSurface:
    return OptilandSurface(
        comment=comment,
        radius=surface.radius,
        thickness=surface.thickness,
        semi_diameter=surface.semi_diameter,
        conic=surface.asphericity,
        material=material,
    )


@make_surface.register
def _make_surface(
    surface: Stop,
    material: Union[str, MaterialModel] = "Air",  # noqa: UP007
    comment: str = "",
) -> OptilandSurface:
    return OptilandSurface(
        comment=comment,
        thickness=surface.thickness,
        material=material,
        semi_diameter=surface.semi_diameter,
        is_stop=True,
    )


@make_surface.register
def _make_surface(
    surface: ZernikeStandardSagSurface,
    material: Union[str, MaterialModel] = "",  # noqa: UP007
    comment: str = "",
) -> OptilandSurface:
    raise NotImplementedError("ZernikeStandardSagSurface is not supported in Optiland.")


@make_surface.register
def _make_surface(
    surface: ZernikeStandardPhaseSurface,
    material: Union[str, MaterialModel] = "",  # noqa: UP007
    comment: str = "",
) -> OptilandSurface:
    raise NotImplementedError("ZernikeStandardPhaseSurface is not supported in Optiland.")


@make_surface.register
def _make_surface(
    surface: NoSurface,  # noqa: ARG001
    material: None = None,  # noqa: ARG001
    comment: str = "",  # noqa: ARG001
) -> OptilandNoSurface:
    return OptilandNoSurface()



from __future__ import annotations

import platform
from typing import TYPE_CHECKING, Any, Literal

import pytest

from visisipy import EyeGeometry, EyeMaterials, EyeModel
from visisipy.models.geometry import NoSurface, StandardSurface, Stop
from visisipy.models.materials import MaterialModel

if platform.system() == "Windows":
    from visisipy.opticstudio import OpticStudioBackend, OpticStudioSettings

from visisipy.optiland import OptilandBackend
from visisipy.optiland.backend import OPTILAND_DEFAULT_SETTINGS

if TYPE_CHECKING:
    from collections.abc import Generator

# Only run the OpticStudio tests on Windows
if platform.system() != "Windows":
    collect_ignore = ["opticstudio"]


def pytest_addoption(parser):
    parser.addoption(
        "--no-opticstudio",
        action="store_true",
        help="Skip tests that require OpticStudio.",
    )
    parser.addoption(
        "--os-extension",
        action="store_true",
        help="Connect to OpticStudio in extension mode.",
    )
    parser.addoption(
        "--os-standalone",
        action="store_false",
        dest="os_extension",
        help="Connect to OpticStudio in standalone mode.",
    )
    parser.addoption("--os-update-ui", action="store_true", help="Show updates in the OpticStudio UI.")


def pytest_collection_modifyitems(config, items):
    for item in items:
        needs_opticstudio = item.get_closest_marker("needs_opticstudio")
        if needs_opticstudio and config.getoption("--no-opticstudio"):
            item.add_marker("skip")


def detect_opticstudio() -> bool:
    if platform.system() != "Windows":
        return False

    import zospy as zp  # noqa: PLC0415

    opticstudio_available: bool = False

    try:
        zos = zp.ZOS()
        opticstudio_available = True

        zos.disconnect()
        assert zos.Application is None

        del zos

    except FileNotFoundError:
        opticstudio_available = False

    del zp
    return opticstudio_available


@pytest.fixture(scope="session")
def opticstudio_connection_mode(request) -> Literal["extension", "standalone"]:
    return "extension" if request.config.getoption("--os-extension") else "standalone"


@pytest.fixture(scope="session")
def opticstudio_available(request) -> bool:
    if request.config.getoption("--no-opticstudio"):
        return False

    return detect_opticstudio()


@pytest.fixture(autouse=True)
def skip_opticstudio(request, opticstudio_available):
    if request.node.get_closest_marker("needs_opticstudio") and not opticstudio_available:
        pytest.skip("OpticStudio is not available.")


@pytest.fixture(autouse=True)
def skip_windows_only(request, opticstudio_available):
    if request.node.get_closest_marker("windows_only") and platform.system() != "Windows":
        pytest.skip("Running on a non-Windows platform.")


@pytest.fixture
def opticstudio_backend(opticstudio_connection_mode, request, opticstudio_available):
    if not opticstudio_available:
        pytest.skip("OpticStudio is not available.")

    if platform.system() != "Windows":
        pytest.skip("Running on a non-Windows platform.")

    OpticStudioBackend.initialize(**OpticStudioSettings(mode=opticstudio_connection_mode))

    if opticstudio_connection_mode == "extension":
        # Disable UI updates using command line option, making the tests run faster
        OpticStudioBackend.zos.Application.ShowChangesInUI = request.config.getoption("--os-update-ui")

    yield OpticStudioBackend

    if OpticStudioBackend.zos is not None and OpticStudioBackend.oss is not None:
        OpticStudioBackend.disconnect()


@pytest.fixture
def optiland_backend() -> Generator[type[OptilandBackend], Any, None]:
    """Fixture to initialize the Optiland backend for testing.

    Returns
    -------
    OptilandBackend
        The initialized Optiland backend.
    """
    OptilandBackend.initialize(**OPTILAND_DEFAULT_SETTINGS)

    yield OptilandBackend

    # Reset settings to defaults
    OptilandBackend.update_settings(**OPTILAND_DEFAULT_SETTINGS)
    OptilandBackend.clear_model()


@pytest.fixture
def eye_model():
    geometry = EyeGeometry(
        cornea_front=StandardSurface(radius=7.72, asphericity=-0.26, thickness=0.55),
        cornea_back=StandardSurface(radius=6.50, asphericity=0, thickness=3.05),
        pupil=Stop(semi_diameter=1.348),
        lens_front=StandardSurface(radius=10.2, asphericity=-3.1316, thickness=4.0),
        lens_back=StandardSurface(radius=-6.0, asphericity=-1, thickness=16.3203),
        retina=StandardSurface(radius=-12.0, asphericity=0),
    )

    materials = EyeMaterials(
        cornea=MaterialModel(refractive_index=1.3777, abbe_number=0, partial_dispersion=0),
        aqueous=MaterialModel(refractive_index=1.3391, abbe_number=0, partial_dispersion=0),
        lens=MaterialModel(refractive_index=1.4222, abbe_number=0, partial_dispersion=0),
        vitreous=MaterialModel(refractive_index=1.3377, abbe_number=0, partial_dispersion=0),
    )

    return EyeModel(geometry=geometry, materials=materials)


@pytest.fixture
def three_surface_eye_model():
    """Eye model with three optical surfaces.

    The cornea back surface is omitted using a `NoSurface` object.
    """
    geometry = EyeGeometry(
        cornea_front=StandardSurface(radius=7.72, asphericity=-0.26, thickness=0.55),
        cornea_back=NoSurface(),
        pupil=Stop(semi_diameter=1.348),
        lens_front=StandardSurface(radius=10.2, asphericity=-3.1316, thickness=4.0),
        lens_back=StandardSurface(radius=-6.0, asphericity=-1, thickness=16.3203),
        retina=StandardSurface(radius=-12.0, asphericity=0),
    )

    return EyeModel(geometry)



from __future__ import annotations

import argparse
import json
import sys
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple

import matplotlib.pyplot as plt
import numpy as np
import zospy as zp
from matplotlib.colors import TABLEAU_COLORS
from matplotlib.lines import Line2D
from optiland.materials import AbbeMaterial
from scipy.optimize import curve_fit

from visisipy.models.materials import (
    NavarroMaterials458,
    NavarroMaterials543,
    NavarroMaterials589,
    NavarroMaterials633,
)

if TYPE_CHECKING:
    from zospy.zpcore import OpticStudioSystem


NAVARRO_MATERIAL_MODELS = {
    0.458: NavarroMaterials458(),
    0.543: NavarroMaterials543(),
    0.5893: NavarroMaterials589(),
    0.6328: NavarroMaterials633(),
}
STRUCTURES = ("cornea", "aqueous", "lens", "vitreous")


def create_model(oss: OpticStudioSystem, *, fit_dispersion: bool = False) -> None:
    oss.new()
    oss.make_sequential()

    oss.SystemData.Fields.SetFieldType(zp.constants.SystemData.FieldType.Angle)
    oss.SystemData.Fields.GetField(1).X = 0
    oss.SystemData.Fields.GetField(1).Y = 0

    assert oss.SystemData.Fields.NumberOfFields == 1, "Number of fields should be 1"

    for i, w in enumerate(NAVARRO_MATERIAL_MODELS.keys()):
        if i == 0:
            oss.SystemData.Wavelengths.GetWavelength(1).Wavelength = w
        else:
            oss.SystemData.Wavelengths.AddWavelength(w, 1)

    assert oss.SystemData.Wavelengths.NumberOfWavelengths == len(NAVARRO_MATERIAL_MODELS), (
        f"Number of wavelengths should be {len(NAVARRO_MATERIAL_MODELS)}"
    )

    oss.SystemData.Aperture.ApertureType = zp.constants.SystemData.ZemaxApertureType.FloatByStopSize

    aperture = oss.LDE.GetSurfaceAt(1)
    aperture.SemiDiameter = 5

    lens_front = oss.LDE.InsertNewSurfaceAt(2)
    lens_front.Thickness = 10
    lens_front.Radius = 10
    material_model = zp.solvers.material_model(
        lens_front.MaterialCell, refractive_index=1, abbe_number=0, partial_dispersion=0
    )
    material_model.VaryIndex = True
    material_model.VaryAbbe = True
    material_model.VarydPgF = fit_dispersion

    lens_front.MaterialCell.SetSolveData(material_model)

    lens_back = oss.LDE.InsertNewSurfaceAt(3)
    lens_back.Thickness = 0
    lens_back.Radius = -10


def _get_wavelength_number(oss: OpticStudioSystem, wavelength: float) -> int:
    for i in range(1, oss.SystemData.Wavelengths.NumberOfWavelengths + 1):
        if oss.SystemData.Wavelengths.GetWavelength(i).Wavelength == wavelength:
            return i

    raise ValueError(f"Wavelength {wavelength} not found in system.")


def build_merit_function(oss: OpticStudioSystem, refractive_indices: dict[float, float]) -> None:
    oss.MFE.ShowEditor()
    oss.MFE.DeleteAllRows()

    comment = oss.MFE.GetOperandAt(1)
    comment.ChangeType(zp.constants.Editors.MFE.MeritOperandType.BLNK)
    comment.GetCellAt(1).Value = "Refractive indices per wavelength"

    for wavelength, refractive_index in refractive_indices.items():
        operand = oss.MFE.InsertNewOperandAt(oss.MFE.NumberOfOperands + 1)
        operand.ChangeType(zp.constants.Editors.MFE.MeritOperandType.INDX)
        operand.GetCellAt(2).Value = str(2)  # Surface number
        operand.GetCellAt(3).Value = str(_get_wavelength_number(oss, wavelength))  # Wavelength number
        operand.Target = refractive_index
        operand.Weight = 1.0


class OptimizationResult(NamedTuple):
    target: float
    value: float

    @property
    def difference(self) -> float:
        return self.value - self.target


class MaterialModel(NamedTuple):
    refractive_index: float
    abbe_number: float = 0
    partial_dispersion: float = 0


def run_optimization(
    oss: OpticStudioSystem,
) -> tuple[MaterialModel, dict[float, OptimizationResult]]:
    if oss.Tools.CurrentTool is not None:
        oss.Tools.CurrentTool.Close()

    local_optimization = oss.Tools.OpenLocalOptimization()
    local_optimization.Algorithm = zp.constants.Tools.Optimization.OptimizationAlgorithm.DampedLeastSquares
    local_optimization.NumberOfCores = 20
    local_optimization.Cycles = zp.constants.Tools.Optimization.OptimizationCycles.Automatic

    local_optimization.RunAndWaitForCompletion()

    if not local_optimization.Succeeded:
        print("Optimization failed", local_optimization.ErrorMessage)
        sys.exit(2)

    local_optimization.Close()

    # oss.MFE.CalculateMeritFunction()

    result = {}

    for i in range(1, oss.MFE.NumberOfOperands + 1):
        operand = oss.MFE.GetOperandAt(i)
        if operand.TypeName == "INDX":
            wavelength = oss.SystemData.Wavelengths.GetWavelength(operand.GetCellAt(3).IntegerValue).Wavelength
            target = operand.Target
            value = operand.Value

            result[wavelength] = OptimizationResult(target, value)

    solve_data = oss.LDE.GetSurfaceAt(2).MaterialCell.GetSolveData()

    return MaterialModel(
        solve_data._S_MaterialModel.IndexNd,  # noqa: SLF001
        solve_data._S_MaterialModel.AbbeVd,  # noqa: SLF001
        solve_data._S_MaterialModel.dPgF,  # noqa: SLF001
    ), result


def fit_opticstudio_model(
    oss: OpticStudioSystem,
    refractive_indices: dict[float, float],
    *,
    fit_dispersion: bool = False,
) -> tuple[MaterialModel, dict[float, OptimizationResult]]:
    create_model(oss, fit_dispersion=fit_dispersion)

    zp.analyses.systemviewers.CrossSection().run(oss)

    build_merit_function(oss, refractive_indices)

    return run_optimization(oss)


def fit_optiland_model(
    refractive_indices: dict[float, float], *, fit_dispersion: bool = False
) -> tuple[MaterialModel, dict[float, OptimizationResult]]:
    if fit_dispersion:
        raise NotImplementedError("Partial dispersion fitting is not supported in OptiLand.")

    def model(wavelength, n, abbe):
        return AbbeMaterial(n, abbe).n(wavelength)

    popt, _ = curve_fit(
        model,
        list(refractive_indices.keys()),
        list(refractive_indices.values()),
        p0=(1, 0),
    )

    fit_model = AbbeMaterial(popt[0], popt[1])

    return MaterialModel(*popt), {w: OptimizationResult(n, fit_model.n(w)) for w, n in refractive_indices.items()}


def opticstudio_calculate_refractive_indices(
    oss: OpticStudioSystem,
    wavelengths: np.ndarray | list[float],
    material_model: MaterialModel,
    surface: int = 2,
) -> np.ndarray:
    create_model(oss)
    zp.solvers.material_model(
        oss.LDE.GetSurfaceAt(surface).MaterialCell,
        refractive_index=material_model.refractive_index,
        abbe_number=material_model.abbe_number,
        partial_dispersion=material_model.partial_dispersion,
    )

    oss.MFE.DeleteAllRows()

    wavelength_number = oss.SystemData.Wavelengths.NumberOfWavelengths + 1
    oss.SystemData.Wavelengths.AddWavelength(wavelengths[0], 1)

    operand = oss.MFE.GetOperandAt(1)
    operand.ChangeType(zp.constants.Editors.MFE.MeritOperandType.INDX)
    operand.GetCellAt(2).Value = str(surface)  # Surface number
    operand.GetCellAt(3).Value = str(wavelength_number)  # Wavelength number

    refractive_indices = np.zeros(len(wavelengths))

    for i, wavelength in enumerate(wavelengths):
        oss.SystemData.Wavelengths.GetWavelength(wavelength_number).Wavelength = wavelength

        oss.MFE.CalculateMeritFunction()
        refractive_indices[i] = operand.Value

    return refractive_indices


def optiland_calculate_refractive_indices(
    wavelengths: np.ndarray | list[float], material_model: MaterialModel
) -> np.ndarray:
    optiland_model = AbbeMaterial(material_model.refractive_index, material_model.abbe_number)

    return np.array([optiland_model.n(wavelength) for wavelength in wavelengths])


def plot_refractive_indices(
    oss: OpticStudioSystem,
    wavelengths: np.ndarray,
    material_models: dict[str, MaterialModel],
) -> None:
    fig, ax = plt.subplots()

    for (structure, material_model), color in zip(material_models.items(), TABLEAU_COLORS, strict=False):
        refractive_indices_opticstudio = opticstudio_calculate_refractive_indices(oss, wavelengths, material_model)
        refractive_indices_optiland = optiland_calculate_refractive_indices(wavelengths, material_model)

        ax.plot(
            wavelengths,
            refractive_indices_opticstudio,
            color=color,
            label=f"{structure}",
        )
        ax.plot(wavelengths, refractive_indices_optiland, color=color, linestyle="--")

        for wavelength, model in NAVARRO_MATERIAL_MODELS.items():
            refractive_index = getattr(model, structure).refractive_index
            ax.scatter([wavelength], [refractive_index], color=color, marker="x")

    ax.set_xlabel("Wavelength [μm]")
    ax.set_ylabel("Refractive index")
    box = ax.get_position()
    ax.set_position((box.x0, box.y0 + box.height * 0.15, box.width, box.height * 0.85))

    handles, labels = ax.get_legend_handles_labels()
    handles.extend(
        [
            Line2D([], [], color="black", label="OpticStudio"),
            Line2D([], [], color="black", linestyle="--", label="Optiland"),
        ]
    )
    labels.extend(
        [
            "OpticStudio",
            "Optiland",
        ]
    )
    ax.legend(handles, labels, loc="upper center", bbox_to_anchor=(0.5, -0.15), ncol=3)

    # plt.tight_layout()
    plt.show()


def save_fit_results(
    material_models: dict[str, MaterialModel],
    optimization_results: dict[str, dict[float, OptimizationResult]],
    filename: str | Path,
) -> None:
    filename = Path(filename).resolve()
    if not filename.parent.exists():
        print(f"Cannot create output file, {filename.parent} does not exist.")
        sys.exit(1)

    result = {
        "models": {
            structure: {
                "refractive_index": model.refractive_index,
                "abbe_number": model.abbe_number,
                "partial_dispersion": model.partial_dispersion,
            }
            for structure, model in material_models.items()
        },
        "mean_absolute_errors": {
            structure: np.mean(np.abs([result.difference for result in results.values()]))
            for structure, results in optimization_results.items()
        },
    }

    filename.write_text(json.dumps(result, indent=4), encoding="utf-8")


def main(args: argparse.Namespace) -> None:
    # Initialize OpticStudio
    zos = zp.ZOS()
    oss = zos.connect("standalone")

    material_models: dict[str, MaterialModel] = {}
    optimization_results: dict[str, dict[float, OptimizationResult]] = {}

    for structure in STRUCTURES:
        refractive_indices = {
            wavelength: getattr(model, structure).refractive_index
            for wavelength, model in NAVARRO_MATERIAL_MODELS.items()
        }

        match args.backend:
            case "optiland":
                material_model, optimization_result = fit_optiland_model(
                    refractive_indices, fit_dispersion=args.fit_dispersion
                )
            case "opticstudio":
                material_model, optimization_result = fit_opticstudio_model(
                    oss, refractive_indices, fit_dispersion=args.fit_dispersion
                )
            case _:
                print(f"Unknown backend: {args.backend}")
                sys.exit(1)

        material_models[structure] = material_model
        optimization_results[structure] = optimization_result

        if args.fit_dispersion:
            print(
                f"{structure}: n = {material_model.refractive_index:.4f}, "
                f"Vd = {material_model.abbe_number:.4f}, "
                f"dPgF = {material_model.partial_dispersion:.4f}"
            )
        else:
            print(f"{structure}: n = {material_model.refractive_index:.4f}, Vd = {material_model.abbe_number:.4f}")

        for wavelength, result in optimization_result.items():
            print(
                f"Wavelength: {wavelength:6f}, Target: {result.target:.4f}, Value: {result.value:.4f}, Difference: {result.difference:+.4g}"
            )

        print("\n")

    if args.output:
        save_fit_results(material_models, optimization_results, args.output)
    else:
        filename = f"material_models_{args.backend}"
        if args.fit_dispersion:
            filename += "_dispersion"

        save_fit_results(
            material_models,
            optimization_results,
            Path(__file__).parent / f"{filename}.json",
        )

    wavelengths = np.linspace(0.4, 0.7, 50)  # μm
    plot_refractive_indices(oss, wavelengths, material_models)


if __name__ == "__main__":
    # Restore default rcparams
    plt.rcdefaults()

    parser = argparse.ArgumentParser(description="Fit material models to refractive indices.")
    parser.add_argument(
        "--backend",
        choices=["optiland", "opticstudio"],
        default="opticstudio",
        help="Backend to use for fitting",
    )
    parser.add_argument(
        "--fit-dispersion",
        action="store_true",
        default=False,
        help="Fit dispersion (only supported in OpticStudio)",
    )
    parser.add_argument("--output", type=Path, default=None, help="Output file for fit results")

    args = parser.parse_args()
    main(args)



"""Analyze optical eye models in Optiland."""

from __future__ import annotations

from visisipy.optiland.backend import OptilandBackend
from visisipy.optiland.models import OptilandEye

__all__ = ("OptilandBackend", "OptilandEye")



"""Spherical equivalent of refraction analysis for Optiland."""

from __future__ import annotations

from typing import TYPE_CHECKING

from visisipy.analysis.refraction import zernike_data_to_refraction
from visisipy.optiland.analysis.zernike_coefficients import (
    zernike_standard_coefficients,
)
from visisipy.types import SampleSize

if TYPE_CHECKING:
    from optiland.wavefront import ZernikeOPD

    from visisipy.backend import FieldCoordinate, FieldType
    from visisipy.optiland.backend import OptilandBackend
    from visisipy.refraction import FourierPowerVectorRefraction

__all__ = ("refraction",)


def refraction(
    backend: type[OptilandBackend],
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    sampling: SampleSize | str | int = 64,
    pupil_diameter: float | None = None,
    field_type: FieldType = "angle",
    *,
    use_higher_order_aberrations: bool = True,
) -> tuple[FourierPowerVectorRefraction, ZernikeOPD]:
    old_aperture = None
    if pupil_diameter is not None:
        old_aperture = backend.get_optic().aperture
        backend.model.pupil.surface.set_semi_aperture(pupil_diameter / 2)

    zernike_coefficients, zernike_opd = zernike_standard_coefficients(
        backend=backend,
        field_coordinate=field_coordinate,
        wavelength=wavelength,
        field_type=field_type,
        sampling=SampleSize(sampling),
    )

    exit_pupil_semi_diameter = backend.get_optic().paraxial.XPD() / 2

    if old_aperture is not None:
        backend.get_optic().set_aperture(aperture_type=old_aperture.ap_type, value=old_aperture.value)

    if wavelength is None:
        wavelength = backend.get_wavelengths()[0]

    fourier_refraction = zernike_data_to_refraction(
        zernike_coefficients,
        exit_pupil_semi_diameter=exit_pupil_semi_diameter,
        wavelength=wavelength,
        use_higher_order_aberrations=use_higher_order_aberrations,
    )

    return fourier_refraction, zernike_opd



"""Types for use throughout visisipy.

This module contains type definitions and utility classes for use throughout the visisipy package.
"""

from __future__ import annotations

import re
import sys
from typing import Literal

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired, TypedDict, Unpack
else:
    from typing import NotRequired, TypedDict, Unpack


__all__ = ("ApertureType", "FieldCoordinate", "FieldType", "NotRequired", "SampleSize", "TypedDict", "Unpack")

ApertureType = Literal[
    "float_by_stop_size",
    "entrance_pupil_diameter",
    "image_f_number",
    "object_numeric_aperture",
]
FieldType = Literal["angle", "object_height"]
FieldCoordinate = tuple[float, float]


RE_SAMPLE_SIZE = re.compile(r"^(?P<sampling>\d+)x(?P=sampling)$", re.IGNORECASE)


class SampleSize:
    """Sample size.

    This class represents a sample size used for various analyses.
    How the sample size is used depends on the backend. Sample sizes can be specified as integers or strings in the
    format "NxN", where N is an integer. Only symmetric sample sizes are supported.

    Attributes
    ----------
    sampling : int
        The sample size.

    Example
    -------
    >>> sample_size = SampleSize(512)
    >>> print(int(sample_size))
    512
    >>> sample_size = SampleSize("512x512")
    >>> print(sample_size)
    512x512
    """

    __slots__ = ("__sampling",)

    def __init__(self, sample_size: int | str | SampleSize):
        """Create a new sample size.

        The sample size can be specified as an integer or a string in the format "NxN", where N is an integer.

        Parameters
        ----------
        sample_size : int | str | SampleSize
            The sample size. Can be an integer, a string in the format "NxN", or a SampleSize object.
        """
        if isinstance(sample_size, int):
            self.__sampling = sample_size

        elif isinstance(sample_size, str):
            match = RE_SAMPLE_SIZE.match(sample_size)

            if not match:
                raise ValueError(f"Invalid sample size format: {sample_size}")

            self.__sampling = int(match.group("sampling"))

        elif isinstance(sample_size, SampleSize):
            self.__sampling = sample_size.sampling

    @property
    def sampling(self) -> int:
        """Get the sample size."""
        return self.__sampling

    def __int__(self):
        return self.sampling

    def __str__(self):
        return f"{self.sampling}x{self.sampling}"

    def __repr__(self):
        return f"SampleSize({self.sampling})"

    def __mul__(self, factor: int) -> SampleSize:
        """Multiply the sample size by an integer.

        Parameters
        ----------
        factor : int
            The integer to multiply the sample size by.

        Returns
        -------
        SampleSize
            A new SampleSize object with the multiplied sample size.
        """
        if not isinstance(factor, int):
            raise TypeError("SampleSize can only be multiplied by integer factors.")
        if factor <= 0:
            raise ValueError("SampleSize factor must be a positive integer.")

        return SampleSize(self.sampling * factor)

    def __rmul__(self, factor: int) -> SampleSize:
        """Multiply the sample size by an integer.

        Parameters
        ----------
        factor : int
            The integer to multiply the sample size by.

        Returns
        -------
        SampleSize
            A new SampleSize object with the multiplied sample size.
        """
        return self.__mul__(factor)



"""Utilities for handling refraction data."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

import numpy as np

__all__ = (
    "FourierPowerVectorRefraction",
    "PolarPowerVectorRefraction",
    "SpheroCylindricalRefraction",
)


@dataclass
class FourierPowerVectorRefraction:
    """Ocular refraction in Fourier power vector form.

    This is the default representation of refractions in visisipy. This class contains several methods to convert the
    refraction to other forms, using the conversions defined in [1]_.

    .. [1] Thibos LN, Wheeler W, Horner D. Power vectors: an application of Fourier analysis to the description and
           statistical analysis of refractive error. Optometry and vision science: official publication of the American
           Academy of Optometry. 1997 Jun;74(6):367-75.

    Attributes
    ----------
    M : float
        The power of the refraction.
    J0 : float
        The Jackson cross-cylinder power at 0°.
    J45 : float
        The Jackson cross-cylinder power at 45°.

    Methods
    -------
    to_polar_power_vectors() -> PolarPowerVectorRefraction
        Converts the refraction to polar power vector form.
    to_sphero_cylindrical(cylinder_form: Literal["positive", "negative"] = "negative") -> SpheroCylindricalRefraction
        Converts the refraction to sphero-cylindrical form.
    """

    M: float
    J0: float
    J45: float

    def to_polar_power_vectors(self) -> PolarPowerVectorRefraction:
        """Converts the refraction to polar power vector form.

        Returns
        -------
        PolarPowerVectorRefraction
            The refraction in polar power vector form.
        """
        return PolarPowerVectorRefraction(
            M=self.M,
            J=np.sqrt(self.J0**2 + self.J45**2),
            axis=np.rad2deg(np.arctan2(self.J45, self.J0) / 2),
        )

    def to_sphero_cylindrical(
        self, cylinder_form: Literal["positive", "negative"] = "negative"
    ) -> SpheroCylindricalRefraction:
        """Converts the refraction to sphero-cylindrical form.

        Parameters
        ----------
        cylinder_form : str, optional
            Indicates if the cylinder should be positive or negative. Defaults to "negative".

        Returns
        -------
        SpheroCylindricalRefraction
            The refraction in sphero-cylindrical form.
        """
        if cylinder_form not in {"positive", "negative"}:
            raise ValueError("cylinder_form must be either 'positive' or 'negative'.")

        sphero_cylinder = SpheroCylindricalRefraction(
            sphere=self.M + np.sqrt(self.J0**2 + self.J45**2),
            cylinder=-2 * np.sqrt(self.J0**2 + self.J45**2),
            axis=np.rad2deg(np.arctan2(self.J45, self.J0) / 2),
        )

        return sphero_cylinder.convert_cylinder_form(cylinder_form)


@dataclass
class PolarPowerVectorRefraction:
    """Ocular refraction in polar power vector form.

    Attributes
    ----------
    M : float
        The power of the refraction.
    J : float
        The Jackson cross-cylinder power.
    axis : float
        The axis of the Jackson cross-cylinder power.
    """

    M: float
    J: float
    axis: float


@dataclass
class SpheroCylindricalRefraction:
    """Ocular refraction in sphero-cylindrical form.

    Attributes
    ----------
    sphere : float
        The spherical component of the refraction.
    cylinder : float
        The cylindrical component of the refraction.
    axis : float
        The axis of the cylindrical component of the refraction.

    Methods
    -------
    has_positive_cylinder() -> bool
        Returns `True` if the cylinder is positive or `NaN`, `False` otherwise.
    has_negative_cylinder() -> bool
        Returns `True` if the cylinder is negative or `NaN`, `False` otherwise.
    convert_cylinder_form(to: Literal["positive", "negative"]) -> SpheroCylindricalRefraction
        Converts from positive cylinder refraction to negative cylinder refraction and vice-versa.
    """

    sphere: float
    cylinder: float
    axis: float

    @property
    def has_positive_cylinder(self) -> bool:
        """Returns `True` if the cylinder is positive or `NaN`, `False` otherwise."""
        return np.isnan(self.cylinder) or self.cylinder >= 0

    @property
    def has_negative_cylinder(self) -> bool:
        """Returns `True` if the cylinder is negative or `NaN`, `False` otherwise."""
        return np.isnan(self.cylinder) or self.cylinder < 0

    def _convert_cylinder_form(self) -> SpheroCylindricalRefraction:
        return SpheroCylindricalRefraction(
            sphere=self.sphere + self.cylinder,
            cylinder=-self.cylinder,
            axis=(self.axis + 90) % 180,
        )

    def convert_cylinder_form(self, to: Literal["positive", "negative"]) -> SpheroCylindricalRefraction:
        """Converts from positive cylinder refraction to negative cylinder refraction and vice-versa.

        Parameters
        ----------
        to : str
            Indicates if the conversion should be done towards negative cylinder form ("negative") or positive cylinder
            form ("positive").

        Returns
        -------
        SpheroCylindricalRefraction, optional
            With inplace set to True, no return is given as the current class is updated. With inplace set to False, a
            copy of the current SpheroCylindricalRefraction instance is returned, but in the specified cylinder format.

        Raises
        ------
        ValueError
            When the parameter "to" is not set to 'positive' or 'negative'.
        """

        if to not in {"negative", "positive"}:
            raise ValueError('"to" should be either "negative" or "positive"')

        if (to == "negative" and self.has_negative_cylinder) or (to == "positive" and self.has_positive_cylinder):
            # Conversion not needed
            return self

        return self._convert_cylinder_form()



from __future__ import annotations

import numpy as np
import pytest

from visisipy.models import EyeGeometry, NavarroGeometry, create_geometry
from visisipy.models.geometry import GeometryParameters, StandardSurface, Stop


@pytest.fixture
def example_geometry_parameters() -> GeometryParameters:
    return {
        "axial_length": 20,
        "cornea_thickness": 0.5,
        "anterior_chamber_depth": 3,
        "lens_thickness": 4,
        "cornea_front_radius": 7,
        "cornea_front_asphericity": 0,
        "cornea_back_radius": 6,
        "cornea_back_asphericity": 0,
        "lens_front_radius": 10,
        "lens_front_asphericity": 0,
        "lens_back_radius": -6,
        "lens_back_asphericity": 0,
        "retina_radius": -12,
        "retina_asphericity": 0,
        "pupil_radius": 1.0,
        "pupil_lens_distance": 0.4,
    }


@pytest.fixture
def example_geometry(example_geometry_parameters):
    return EyeGeometry(
        cornea_front=StandardSurface(
            radius=example_geometry_parameters["cornea_front_radius"],
            asphericity=example_geometry_parameters["cornea_front_asphericity"],
            thickness=example_geometry_parameters["cornea_thickness"],
        ),
        cornea_back=StandardSurface(
            radius=example_geometry_parameters["cornea_back_radius"],
            asphericity=example_geometry_parameters["cornea_back_asphericity"],
            thickness=example_geometry_parameters["anterior_chamber_depth"],
        ),
        pupil=Stop(
            semi_diameter=example_geometry_parameters["pupil_radius"],
            thickness=example_geometry_parameters["pupil_lens_distance"],
        ),
        lens_front=StandardSurface(
            radius=example_geometry_parameters["lens_front_radius"],
            asphericity=example_geometry_parameters["lens_front_asphericity"],
            thickness=example_geometry_parameters["lens_thickness"],
        ),
        lens_back=StandardSurface(
            radius=example_geometry_parameters["lens_back_radius"],
            asphericity=example_geometry_parameters["lens_back_asphericity"],
            thickness=round(
                example_geometry_parameters["axial_length"]
                - example_geometry_parameters["cornea_thickness"]
                - example_geometry_parameters["anterior_chamber_depth"]
                - example_geometry_parameters["pupil_lens_distance"]
                - example_geometry_parameters["lens_thickness"],
                7,
            ),
        ),
        retina=StandardSurface(
            radius=example_geometry_parameters["retina_radius"],
            asphericity=example_geometry_parameters["retina_asphericity"],
        ),
    )


class TestStandardSurface:
    @pytest.mark.parametrize(
        "radius,asphericity,z_radius,y_radius",
        [
            (12, 0, 12, 12),
            (12, 1, 6, 12 / np.sqrt(2)),
            (12, -0.5, 24, 12 / np.sqrt(0.5)),
        ],
    )
    def test_ellipsoid_radii(self, radius, asphericity, z_radius, y_radius):
        surface = StandardSurface(radius=radius, asphericity=asphericity)

        assert surface.ellipsoid_radii == pytest.approx((z_radius, y_radius, y_radius))
        assert surface.ellipsoid_radii.z == pytest.approx(z_radius)
        assert surface.ellipsoid_radii.y == pytest.approx(y_radius)
        assert surface.ellipsoid_radii.x == pytest.approx(y_radius)

    @pytest.mark.parametrize("asphericity", [-1, -1.5])
    def test_ellipsoid_radii_raises_notimplementederror(self, asphericity):
        surface = StandardSurface(radius=12, asphericity=asphericity)

        with pytest.raises(NotImplementedError):
            _ = surface.ellipsoid_radii


class TestStop:
    def test_init_stop(self):
        stop = Stop()

        assert stop.is_stop
        assert stop.semi_diameter > 0


class TestEyeGeometry:
    def test_init_geometry(self):
        assert EyeGeometry(
            cornea_front=StandardSurface(),
            cornea_back=StandardSurface(),
            pupil=Stop(),
            lens_front=StandardSurface(),
            lens_back=StandardSurface(),
            retina=StandardSurface(),
        )

    def test_init_geometry_non_stop_pupil_raises_valueerror(self):
        with pytest.raises(ValueError, match="The pupil surface must be a stop"):
            EyeGeometry(
                cornea_front=StandardSurface(),
                cornea_back=StandardSurface(),
                pupil=StandardSurface(is_stop=False),
                lens_front=StandardSurface(),
                lens_back=StandardSurface(),
                retina=StandardSurface(),
            )

    @pytest.mark.parametrize("asphericity", [-1, -1.5])
    def test_init_geometry_non_ellipsoid_retina_raises_valueerror(self, asphericity):
        with pytest.raises(ValueError, match="Only an elliptical retina is allowed"):
            EyeGeometry(
                cornea_front=StandardSurface(),
                cornea_back=StandardSurface(),
                pupil=Stop(),
                lens_front=StandardSurface(),
                lens_back=StandardSurface(),
                retina=StandardSurface(radius=12, asphericity=asphericity),
            )

    def test_axial_length(self, example_geometry_parameters, example_geometry):
        assert example_geometry.axial_length == example_geometry_parameters["axial_length"]

    def test_cornea_thickness(self, example_geometry_parameters, example_geometry):
        assert example_geometry.cornea_thickness == example_geometry_parameters["cornea_thickness"]

    def test_anterior_chamber_depth(self, example_geometry_parameters, example_geometry):
        assert example_geometry.anterior_chamber_depth == example_geometry_parameters["anterior_chamber_depth"]

    def test_pupil_lens_distance(self, example_geometry_parameters, example_geometry):
        assert example_geometry.pupil_lens_distance == example_geometry_parameters["pupil_lens_distance"]

    def test_lens_thickness(self, example_geometry_parameters, example_geometry):
        assert example_geometry.lens_thickness == example_geometry_parameters["lens_thickness"]

    def test_vitreous_thickness(self, example_geometry_parameters, example_geometry):
        vitreous_thickness = example_geometry_parameters["axial_length"] - (
            example_geometry_parameters["cornea_thickness"]
            + example_geometry_parameters["anterior_chamber_depth"]
            + example_geometry_parameters["pupil_lens_distance"]
            + example_geometry_parameters["lens_thickness"]
        )
        assert example_geometry.vitreous_thickness == vitreous_thickness


@pytest.mark.parametrize("base_geometry", [NavarroGeometry])
class TestCreateGeometry:
    def test_create_geometry(self, base_geometry, example_geometry_parameters, example_geometry):
        class SentinelFloat(float):
            """Custom float class to mark floats as set by a unit test."""

        geometry = create_geometry(
            base=base_geometry,
            **{k: SentinelFloat(v) for k, v in example_geometry_parameters.items()},
            estimate_cornea_back=False,
        )

        assert geometry.__dict__ == example_geometry.__dict__

        assert isinstance(geometry.cornea_front.thickness, SentinelFloat)
        assert isinstance(geometry.cornea_front.radius, SentinelFloat)
        assert isinstance(geometry.cornea_front.asphericity, SentinelFloat)

        assert isinstance(geometry.cornea_back.radius, SentinelFloat)
        assert isinstance(geometry.cornea_back.asphericity, SentinelFloat)
        assert isinstance(geometry.cornea_back.thickness, SentinelFloat)

        assert isinstance(geometry.pupil.semi_diameter, SentinelFloat)
        assert isinstance(geometry.pupil.thickness, SentinelFloat)

        assert isinstance(geometry.lens_front.thickness, SentinelFloat)
        assert isinstance(geometry.lens_front.radius, SentinelFloat)
        assert isinstance(geometry.lens_front.asphericity, SentinelFloat)

        # Note: the lens back thickness is a calculated value, so it is not a SentinelFloat
        assert isinstance(geometry.lens_back.radius, SentinelFloat)
        assert isinstance(geometry.lens_back.asphericity, SentinelFloat)

        assert isinstance(geometry.retina.radius, SentinelFloat)
        assert isinstance(geometry.retina.asphericity, SentinelFloat)

    def test_create_geometry_estimate_cornea_back(self, base_geometry, example_geometry_parameters):
        with pytest.warns(
            UserWarning,
            match="The cornea back radius was provided, but it will be ignored",
        ):
            geometry = create_geometry(
                base=base_geometry,
                **example_geometry_parameters,
                estimate_cornea_back=True,
            )

        assert geometry.cornea_back.radius == 0.81 * geometry.cornea_front.radius

    def test_base_no_class_raises_typeerror(self, base_geometry):
        with pytest.raises(TypeError, match="The base geometry must be a class."):
            create_geometry(base=base_geometry())

    def test_base_no_eyegeometry_raises_typeerror(self, base_geometry):
        class InvalidGeometry: ...

        with pytest.raises(TypeError, match="The base geometry must be a subclass of EyeGeometry."):
            create_geometry(base=InvalidGeometry)

    def test_estimate_cornea_back_and_radius_warns(self, base_geometry, example_geometry_parameters):
        with pytest.warns(
            UserWarning,
            match="The cornea back radius was provided, but it will be ignored because estimate_cornea_back is True",
        ):
            create_geometry(
                base=base_geometry,
                estimate_cornea_back=True,
                **example_geometry_parameters,
            )

    @pytest.mark.parametrize(
        "parameters_a",
        [
            {"retina_radius": 12},
            {"retina_asphericity": 0},
            {"retina_radius": 12, "retina_asphericity": 0},
        ],
    )
    @pytest.mark.parametrize(
        "parameters_b",
        [
            {"retina_ellipsoid_z_radius": 12},
            {"retina_ellipsoid_y_radius": 12},
            {"retina_ellipsoid_z_radius": 12, "retina_ellipsoid_y_radius": 12},
        ],
    )
    def test_supplying_retina_parameters_and_half_axes_raises_valueerror(
        self, base_geometry, parameters_a, parameters_b
    ):
        parameters = parameters_a | parameters_b

        with pytest.raises(ValueError, match="Cannot specify both retina radius/asphericity and ellipsoid radii"):
            create_geometry(base=base_geometry, **parameters)

    @pytest.mark.parametrize(
        "parameters",
        [
            {"retina_ellipsoid_z_radius": 12},
            {"retina_ellipsoid_y_radius": 12},
        ],
    )
    def test_supplying_single_half_axis_raises_valueerror(self, base_geometry, parameters):
        with pytest.raises(
            ValueError,
            match="If the retina ellipsoid radii are specified, both the z and y radius must be provided",
        ):
            create_geometry(base=base_geometry, **parameters)

    def test_incomplete_thickness_raises_valueerror(self, base_geometry, monkeypatch):
        # Mock the __init__ method to avoid setting the thickness of the cornea front
        def mock_init(self, *args, **kwargs):
            self.cornea_front = StandardSurface(thickness=None)
            self.cornea_back = StandardSurface()
            self.pupil = Stop()
            self.lens_front = StandardSurface()
            self.lens_back = StandardSurface()
            self.retina = StandardSurface()

        monkeypatch.setattr(base_geometry, "__init__", mock_init)

        with pytest.raises(
            ValueError,
            match="Cannot calculate vitreous thickness from the supplied parameters.",
        ):
            create_geometry(base_geometry, axial_length=20)

    @pytest.mark.parametrize(
        "geometry_parameters",
        [
            {"axial_length": 10, "cornea_thickness": 1, "anterior_chamber_depth": 4, "lens_thickness": 5},
            {"axial_length": 10, "cornea_thickness": 2, "anterior_chamber_depth": 4, "lens_thickness": 5},
        ],
    )
    def test_invalid_vitreous_thickness_raises_valueerror(self, base_geometry, geometry_parameters):
        with pytest.raises(
            ValueError,
            match="The sum of the cornea thickness, anterior chamber depth, pupil-lens distance and lens thickness is "
            "greater than or equal to the axial length.",
        ):
            create_geometry(
                base=base_geometry,
                axial_length=10,
                cornea_thickness=1,
                anterior_chamber_depth=4,
                lens_thickness=5,
            )

    def test_set_retina_ellipsoid_radii(self, base_geometry):
        geometry = create_geometry(base=base_geometry, retina_ellipsoid_z_radius=12.34, retina_ellipsoid_y_radius=10)

        assert pytest.approx(geometry.retina.ellipsoid_radii.z) == 12.34
        assert pytest.approx(geometry.retina.ellipsoid_radii.y) == 10
        assert pytest.approx(geometry.retina.ellipsoid_radii.x) == 10



"""Cardinal points analysis for OpticStudio."""

from __future__ import annotations

from typing import TYPE_CHECKING

import zospy as zp

from visisipy.analysis.cardinal_points import CardinalPoints, CardinalPointsResult

if TYPE_CHECKING:
    from zospy.analyses.reports.cardinal_points import CardinalPointsResult as ZOSPyCardinalPointsResult

    from visisipy.opticstudio.backend import OpticStudioBackend


def _build_cardinal_points_result(cardinal_points_result: ZOSPyCardinalPointsResult) -> CardinalPointsResult:
    return CardinalPointsResult(
        focal_lengths=CardinalPoints(
            image=cardinal_points_result.cardinal_points.focal_length.image,
            object=cardinal_points_result.cardinal_points.focal_length.object,
        ),
        focal_points=CardinalPoints(
            image=cardinal_points_result.cardinal_points.focal_planes.image,
            object=cardinal_points_result.cardinal_points.focal_planes.object,
        ),
        principal_points=CardinalPoints(
            image=cardinal_points_result.cardinal_points.principal_planes.image,
            object=cardinal_points_result.cardinal_points.principal_planes.object,
        ),
        anti_principal_points=CardinalPoints(
            image=cardinal_points_result.cardinal_points.anti_principal_planes.image,
            object=cardinal_points_result.cardinal_points.anti_principal_planes.object,
        ),
        nodal_points=CardinalPoints(
            image=cardinal_points_result.cardinal_points.nodal_planes.image,
            object=cardinal_points_result.cardinal_points.nodal_planes.object,
        ),
        anti_nodal_points=CardinalPoints(
            image=cardinal_points_result.cardinal_points.anti_nodal_planes.image,
            object=cardinal_points_result.cardinal_points.anti_nodal_planes.object,
        ),
    )


def cardinal_points(
    backend: OpticStudioBackend, surface_1: int | None = None, surface_2: int | None = None
) -> tuple[CardinalPointsResult, ZOSPyCardinalPointsResult]:
    """Get the cardinal points of the system between `surface_1` and `surface_2`.

    Parameters
    ----------
    backend : type[OpticStudioBackend]
        Reference to the OpticStudio backend.
    surface_1 : int | None, optional
        The first surface to be used in the analysis. If `None`, the first surface in the system will be used.
        Defaults to `None`.
    surface_2 : int | None, optional
        The second surface to be used in the analysis. If `None`, the last surface in the system will be used.
        Defaults to `None`.

    Returns
    -------
    CardinalPointsResult
        The cardinal points of the system.

    Raises
    ------
    ValueError
        If `surface_1` or `surface_2` are not between 1 and the number of surfaces in the system, or if `surface_1`
        is greater than or equal to `surface_2`.
    """
    surface_1 = surface_1 or 1
    surface_2 = surface_2 or backend.get_oss().LDE.NumberOfSurfaces - 1

    if surface_1 < 1 or surface_2 > backend.get_oss().LDE.NumberOfSurfaces - 1:
        raise ValueError("surface_1 and surface_2 must be between 1 and the number of surfaces in the system.")

    if surface_1 >= surface_2:
        raise ValueError("surface_1 must be less than surface_2.")

    cardinal_points_result = zp.analyses.reports.CardinalPoints(
        surface_1=surface_1,
        surface_2=surface_2,
    ).run(backend.get_oss())

    return _build_cardinal_points_result(cardinal_points_result.data), cardinal_points_result.data



from __future__ import annotations

from contextlib import nullcontext as does_not_raise

import pytest

from visisipy.wavefront import ZernikeCoefficients, min_max_noll_index


class TestZernikeCoefficients:
    @pytest.mark.parametrize(
        "terms,result,expectation",
        [
            (None, {}, does_not_raise()),
            ({}, {}, does_not_raise()),
            ({1: 0.5, 2: 0.9}, {1: 0.5, 2: 0.9}, does_not_raise()),
            ({1: 0.5, "x": 0.9}, {}, pytest.raises(TypeError, match="All keys must be integers")),
            ({1: 0.5, 2.8: 0.9}, {}, pytest.raises(TypeError, match="All keys must be integers")),
            ({1: 0.5, 0: 0.9}, {}, pytest.raises(ValueError, match="The Zernike coefficients must be larger than 0")),
            ({-1: 0.5, 2: 0.9}, {}, pytest.raises(ValueError, match="The Zernike coefficients must be larger than 0")),
        ],
    )
    def test_init(self, terms, result, expectation):
        with expectation:
            zernike = ZernikeCoefficients(terms)
            assert zernike == result

    @pytest.mark.parametrize(
        "n,value,expectation",
        [
            (1, 1.2, does_not_raise()),
            (1, 0.0, does_not_raise()),
            (0, 0.0, pytest.raises(ValueError, match="The coefficient must be larger than 0")),
            (-1, 0.0, pytest.raises(ValueError, match="The coefficient must be larger than 0")),
            ("x", 0.0, pytest.raises(TypeError, match="The coefficient must be an integer or a tuple of two integers")),
            (1.5, 0.0, pytest.raises(TypeError, match="The coefficient must be an integer or a tuple of two integers")),
        ],
    )
    def test_setitem(self, n: int, value: float, expectation):
        zernike = ZernikeCoefficients()

        with expectation:
            zernike[n] = value
            assert zernike[n] == value

    @pytest.mark.parametrize(
        "n,expected_value,expectation",
        [
            (1, 0.5, does_not_raise()),
            ((0, 0), 0.5, does_not_raise()),
            (2, 0.9, does_not_raise()),
            ((1, 1), 0.9, does_not_raise()),
            (3, 0.0, does_not_raise()),
            ((1, -1), 0.0, does_not_raise()),
            (0, 0.0, pytest.raises(ValueError, match="The coefficient must be larger than 0")),
            (-1, 0.0, pytest.raises(ValueError, match="The coefficient must be larger than 0")),
            ("x", 0.0, pytest.raises(TypeError, match="The coefficient must be an integer or a tuple of two integers")),
            (1.5, 0.0, pytest.raises(TypeError, match="The coefficient must be an integer or a tuple of two integers")),
            ((-1, 0), 0.0, pytest.raises(ValueError, match="n must be greater than or equal to 0")),
            ((0, 1), 0.0, pytest.raises(ValueError, match="m must be less than or equal to n")),
            ((3, 5), 0.0, pytest.raises(ValueError, match="m must be less than or equal to n")),
            ((5, -4), 0.0, pytest.raises(ValueError, match="n and m must have the same parity")),
            ((5, 4), 0.0, pytest.raises(ValueError, match="n and m must have the same parity")),
        ],
    )
    def test_getitem(self, n, expected_value, expectation):
        zernike = ZernikeCoefficients({1: 0.5, 2: 0.9})

        with expectation:
            assert zernike[n] == expected_value

    @pytest.mark.parametrize(
        "nm,noll",
        [
            ((0, 0), 1),
            ((1, 1), 2),
            ((1, -1), 3),
            ((2, 0), 4),
            ((2, -2), 5),
            ((2, 2), 6),
            ((3, -1), 7),
            ((3, 1), 8),
            ((3, -3), 9),
            ((3, 3), 10),
            ((4, 0), 11),
            ((4, 2), 12),
            ((4, -2), 13),
            ((4, 4), 14),
            ((4, -4), 15),
            ((5, 1), 16),
            ((5, -1), 17),
            ((5, 3), 18),
            ((5, -3), 19),
            ((5, 5), 20),
        ],
    )
    def test_to_noll(self, nm, noll):
        assert ZernikeCoefficients.to_noll(*nm) == noll


class TestMinMaxNollIndex:
    @pytest.mark.parametrize("order", range(100))
    def test_consistent_with_to_noll(self, order: int):
        min_index, max_index = min_max_noll_index(order, order)

        assert min_index == min(ZernikeCoefficients.to_noll(order, s * (order % 2)) for s in (-1, 1))
        assert max_index == max(ZernikeCoefficients.to_noll(order, s * order) for s in (-1, 1))

    @pytest.mark.parametrize("order", range(100))
    def test_min_max_adjacent(self, order: int):
        min_index, max_index = min_max_noll_index(order + 1, order)

        assert min_index == max_index + 1



"""Optiland backend for Visisipy."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, cast

import optiland.backend
from optiland.fileio import save_optiland_file
from optiland.optic import Optic

from visisipy.backend import BackendSettings, BaseBackend, Unpack, _classproperty
from visisipy.optiland.analysis import OptilandAnalysisRegistry
from visisipy.optiland.models import OptilandEye

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable
    from os import PathLike

    from visisipy import EyeModel


__all__ = (
    "OptilandBackend",
    "OptilandSettings",
)


ComputationBackend = Literal["numpy", "torch"]
TorchDevice = Literal["cpu", "cuda"]
TorchPrecision = Literal["float32", "float64"]


class OptilandSettings(BackendSettings, total=False):
    """Backend settings that are specific to the Optiland backend."""

    computation_backend: ComputationBackend
    """Backend used for numerical operations. Must be one of 'numpy' or 'torch'. See
    https://optiland.readthedocs.io/en/latest/developers_guide/configurable_backend.html for more details.

    When using 'torch', the 'torch' package must be installed manually.
    """

    torch_device: TorchDevice
    """Device to use for torch backend. Must be one of 'cpu' or 'cuda'. Only used if computation_backend is 'torch'."""

    torch_precision: TorchPrecision
    """Precision to use for torch backend. Must be one of 'float32' or 'float64'.
    Only used if computation_backend is 'torch'.
    """

    torch_use_grad_mode: bool
    """Globally enable or disable autodifferentiation for the 'torch' backend. Only used if computation_backend is
    'torch'.
    """


OPTILAND_DEFAULT_SETTINGS: OptilandSettings = {
    "field_type": "angle",
    "fields": [(0, 0)],
    "wavelengths": [0.543],
    "aperture_type": "entrance_pupil_diameter",
    "aperture_value": 1,
    "computation_backend": "numpy",
    "torch_device": "cpu",
    "torch_precision": "float32",
    "torch_use_grad_mode": False,
}
"""Default settings for the Optiland backend."""

OPTILAND_APERTURES = {
    "float_by_stop_size": NotImplemented,
    "entrance_pupil_diameter": "EPD",
    "image_f_number": "imageFNO",
    "object_numeric_aperture": "objectNA",
}


class OptilandBackend(BaseBackend):
    """Optiland backend."""

    optic: Optic | None = None
    model: OptilandEye | None = None
    settings: OptilandSettings = OptilandSettings(**OPTILAND_DEFAULT_SETTINGS)
    _analysis: OptilandAnalysisRegistry | None = None

    @_classproperty
    def analysis(cls) -> OptilandAnalysisRegistry:  # noqa: N805
        """Provides access to the `OptilandAnalysisRegistry` instance.

        This property provides access to the `OptilandAnalysisRegistry` instance for performing various analyses on the optical
        system.

        Returns
        -------
        OpticStudioAnalysisRegistry
            The `OptilandAnalysisRegistry` instance.

        Raises
        ------
        RuntimeError
            If the Optiland backend has not been initialized.
        """
        if cls._analysis is None:
            cls._analysis = OptilandAnalysisRegistry(cls)

        return cls._analysis

    @classmethod
    def initialize(cls, **settings: Unpack[OptilandSettings]) -> None:
        """Initialize the Optiland backend.

        This method initializes the Optiland backend with the given settings and creates a new model.

        Parameters
        ----------
        settings : OptilandSettings | None, optional
            The settings to be used for the Optiland backend. If None, the default settings are used.
        """
        if len(settings) > 0:
            cls.settings.update(settings)

        cls.new_model()

    @classmethod
    def update_settings(cls, **settings: Unpack[OptilandSettings]) -> None:
        """Apply the provided settings to the Optiland backend.

        This method applies the provided settings to the Optiland backend.
        """
        if len(settings) > 0:
            cls.settings.update(settings)

        if cls.optic is not None:
            cls.set_aperture()
            cls.set_fields(
                field_type=cls.get_setting("field_type"),
                coordinates=cls.get_setting("fields"),
            )
            cls.set_wavelengths(cls.get_setting("wavelengths"))
            cls.set_computation_backend(
                cls.get_setting("computation_backend"),
                torch_device=cls.get_setting("torch_device"),
                torch_precision=cls.get_setting("torch_precision"),
                torch_use_grad_mode=cls.get_setting("torch_use_grad_mode"),
            )

    @classmethod
    def new_model(
        cls,
        *,
        save_old_model: bool = False,
        save_filename: PathLike | str | None = None,
    ) -> None:
        """Initialize a new optical system.

        This method initializes a new, empty optical system.
        """
        if save_old_model:
            cls.save_model(save_filename)

        cls.optic = Optic()

        cls.update_settings()

    @classmethod
    def build_model(
        cls,
        model: EyeModel,
        *,
        start_from_index: int = 0,
        replace_existing: bool = False,
        object_distance: float = float("inf"),
        **kwargs,
    ) -> OptilandEye:
        """Builds an optical system based on the provided eye model.

        This method creates an OptilandEye instance from the provided eye model and builds the optical system.
        If `replace_existing` is True, any existing model is updated instead of building a completely new system.

        Parameters
        ----------
        model : EyeModel
            The eye model to be used for building the optical system model.
        start_from_index : int, optional
            The index of the surface after which the eye model will be built. The cornea front surface will be located
            at `start_from_index + 1`.
        replace_existing : bool, optional
            Whether to replace any existing model before building the new one. Defaults to False.
        object_distance : float
            Distance between the cornea front and the surface preceding the eye model.
        **kwargs
            Additional keyword arguments to be passed to the OptilandEye build method.

        Returns
        -------
        OptilandEye
            The built optical system model.
        """
        if not replace_existing and cls.model is not None:
            cls.new_model()

        optiland_eye = OptilandEye(model)
        optiland_eye.build(
            cls.get_optic(),
            start_from_index=start_from_index,
            replace_existing=replace_existing,
            object_distance=object_distance,
            **kwargs,
        )

        cls.model = optiland_eye
        return optiland_eye

    @classmethod
    def clear_model(cls) -> None:
        """Clear the current optical system model.

        This method initializes a new optical system, discarding any existing model.
        """
        cls.model = None
        cls.new_model()

    @classmethod
    def save_model(cls, path: str | PathLike | None = None) -> None:
        """Save the current optical system model.

        This method saves the current optical system model to the specified path. If no path is provided,
        it saves the model to the current working directory with the default name (model.json).

        Parameters
        ----------
        path : str | PathLike | None, optional
            The path where the model should be saved. If None, the model is saved in the current working directory.
            The file extension must be .json.
        """
        if path is None:
            path = "model.json"

        if not str(path).endswith(".json"):
            raise ValueError("filename must end in .json")

        save_optiland_file(cls.optic, path)

    @classmethod
    def get_optic(cls) -> Optic:
        """Get the optic object.

        Returns
        -------
        Optic
            The optic object.

        Raises
        ------
        RuntimeError
            If the optic object is not initialized.
        """
        if cls.optic is None:
            raise RuntimeError("No optic object initialized. Please initialize the backend first.")

        return cast(Optic, cls.optic)

    @classmethod
    def set_aperture(cls):
        aperture_type = cls.get_setting("aperture_type")
        aperture_value = cls.get_setting("aperture_value")

        if aperture_type not in OPTILAND_APERTURES:
            raise ValueError(
                f"Invalid aperture type '{aperture_type}'. Must be one of {list(OPTILAND_APERTURES.keys())}."
            )

        if OPTILAND_APERTURES[aperture_type] is NotImplemented:
            raise NotImplementedError(f"Aperture type '{aperture_type}' is not implemented in Optiland.")

        cls.get_optic().set_aperture(
            aperture_type=OPTILAND_APERTURES[aperture_type],
            value=aperture_value,
        )

    @classmethod
    def get_fields(cls) -> list[tuple[float, float]]:
        """Get the fields in the optical system.

        Returns
        -------
        list[tuple[float, float]]
            List of field coordinates.
        """
        return [(f.x, f.y) for f in cls.get_optic().fields.fields]

    @classmethod
    def set_fields(
        cls,
        coordinates: Iterable[tuple[float, float]],
        field_type: Literal["angle", "object_height"] = "angle",
    ):
        """Set the fields for the optical system.

        This method removes any existing fields and adds the new ones provided.

        Parameters
        ----------
        coordinates : Iterable[tuple[float, float]]
            An iterable of tuples representing the coordinates for the fields.
        field_type : Literal["angle", "object_height"], optional
            The type of field to be used in the optical system. Can be either "angle" or "object_height".
            Defaults to "angle".
        """
        if field_type not in {"angle", "object_height"}:
            raise ValueError("field_type must be either 'angle' or 'object_height'.")

        # Remove all fields
        cls.get_optic().fields.fields.clear()

        cls.get_optic().set_field_type(field_type)

        for field in coordinates:
            cls.get_optic().add_field(y=field[1], x=field[0])

    @classmethod
    def get_wavelengths(cls) -> list[float]:
        """Get the wavelengths in the optical system.

        Returns
        -------
        list[float]
            List of wavelengths.
        """
        return [w.value for w in cls.get_optic().wavelengths.wavelengths]

    @classmethod
    def set_wavelengths(cls, wavelengths: Iterable[float]):
        """Set the wavelengths for the optical system.

        This method removes any existing wavelengths and adds the new ones provided.
        The weight for each wavelength is set to 1.0.

        Parameters
        ----------
        wavelengths : Iterable[float]
            An iterable of wavelengths to be set for the optical system.
        """
        # Remove all wavelengths
        cls.get_optic().wavelengths.wavelengths.clear()

        for wavelength in wavelengths:
            cls.get_optic().add_wavelength(wavelength)

    @classmethod
    def iter_fields(cls) -> Generator[tuple[int, tuple[float, float]], Any, None]:
        """Iterate over the fields in the optical system.

        Yields
        ------
        int
            Field index.
        tuple[float, float]
            Field X and Y coordinates.
        """
        for i, f in enumerate(cls.get_optic().fields.fields):
            yield i, (f.x, f.y)

    @classmethod
    def iter_wavelengths(cls) -> Generator[tuple[int, float], Any, None]:
        """Iterate over the wavelengths in the optical system.

        Yields
        ------
        int
            Wavelength index.
        float
            Wavelength value.
        """
        for i, w in enumerate(cls.get_optic().wavelengths.wavelengths):
            yield i, w.value

    @classmethod
    def set_computation_backend(
        cls,
        backend: ComputationBackend,
        *,
        torch_device: TorchDevice = "cpu",
        torch_precision: TorchPrecision = "float32",
        torch_use_grad_mode: bool = False,
    ) -> None:
        """Set the computation backend for Optiland.

        Parameters
        ----------
        backend : ComputationBackend
            The computation backend to use. Must be one of 'numpy' or 'torch'.
        torch_device : str, optional
            The device to use for the 'torch' backend. Must be one of 'cpu' or 'cuda'. Defaults to 'cpu'.
        torch_precision : str, optional
            The precision to use for the 'torch' backend. Must be one of 'float32' or 'float64'. Defaults to 'float32'.
        torch_use_grad_mode : bool, optional
            Whether to enable gradient mode for the 'torch' backend. Defaults to False.
        """
        if backend not in {"numpy", "torch"}:
            raise ValueError("computation_backend must be either 'numpy' or 'torch'.")

        if torch_device not in {"cpu", "cuda"}:
            raise ValueError("torch_device must be either 'cpu' or 'cuda'.")

        if torch_precision not in {"float32", "float64"}:
            raise ValueError("torch_precision must be either 'float32' or 'float64'.")

        if optiland.backend.get_backend() != backend:
            optiland.backend.set_backend(backend)

        if backend == "torch":
            if optiland.backend.get_device() != torch_device:
                optiland.backend.set_device(torch_device)
            if optiland.backend.get_precision() != torch_precision:
                optiland.backend.set_precision(torch_precision)
            if torch_use_grad_mode and not optiland.backend.grad_mode.requires_grad:
                optiland.backend.grad_mode.enable()
            elif not torch_use_grad_mode and optiland.backend.grad_mode.requires_grad:
                optiland.backend.grad_mode.disable()



from __future__ import annotations

import pytest
import zospy as zp


@pytest.fixture
def zos():
    return zp.ZOS.get_instance() or zp.ZOS()


@pytest.fixture
def oss(zos, opticstudio_connection_mode, request):
    oss = zos.connect(mode=opticstudio_connection_mode) if zos.Application is None else zos.get_primary_system()

    if opticstudio_connection_mode == "extension":
        # Disable UI updates using command line option, making the tests run faster
        zos.Application.ShowChangesInUI = request.config.getoption("--os-update-ui")

    oss.new()

    yield oss

    zos.disconnect()



from __future__ import annotations

from contextlib import nullcontext as does_not_raise

import pytest

from visisipy.types import SampleSize


class TestSampleSize:
    @pytest.mark.parametrize(
        "input_,value,expectation",
        [
            (32, 32, does_not_raise()),
            ("64x64", 64, does_not_raise()),
            ("512X512", 512, does_not_raise()),
            (SampleSize(128), 128, does_not_raise()),
            ("123", 123, pytest.raises(ValueError, match="Invalid sample size format: 123")),
            ("123x456", 123, pytest.raises(ValueError, match="Invalid sample size format: 123x456")),
        ],
    )
    def test_samplesize(self, input_, value, expectation):
        with expectation:
            sampling = SampleSize(input_)
            assert sampling.sampling == value

    def test_int_conversion(self):
        assert int(SampleSize(32)) == 32

    def test_str_conversion(self):
        assert str(SampleSize(32)) == "32x32"



"""Ray trace analysis for OpticStudio."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd
import zospy as zp

if TYPE_CHECKING:
    from collections.abc import Iterable

    from zospy.analyses.raysandspots.single_ray_trace import SingleRayTraceResult

    from visisipy.opticstudio.backend import OpticStudioBackend
    from visisipy.types import FieldCoordinate, FieldType


def _build_raytrace_result(raytrace_results: list[pd.DataFrame]) -> pd.DataFrame:
    columns = {
        "Field": "field",
        "Wavelength": "wavelength",
        "Surf": "surface",
        "Comment": "comment",
        "X-coordinate": "x",
        "Y-coordinate": "y",
        "Z-coordinate": "z",
    }

    return pd.concat(raytrace_results)[columns.keys()].rename(columns=columns).reset_index()


def raytrace(
    backend: type[OpticStudioBackend],
    coordinates: Iterable[FieldCoordinate] | None = None,
    wavelengths: Iterable[float] | None = None,
    field_type: FieldType = "angle",
    pupil: tuple[float, float] = (0, 0),
) -> tuple[pd.DataFrame, list[SingleRayTraceResult]]:
    """Perform a ray trace analysis using the given parameters.
    The ray trace is performed for each wavelength and field in the system, using the Single Ray Trace analysis
    in OpticStudio.

    The analysis returns a Dataframe with the following columns:

    - field: The field coordinates for the ray trace.
    - wavelength: The wavelength used in the ray trace.
    - surface: The surface number in the system.
    - comment: The comment for the surface.
    - x: The X-coordinate of the ray trace.
    - y: The Y-coordinate of the ray trace.
    - z: The Z-coordinate of the ray trace.

    Parameters
    ----------
    backend : type[OpticStudioBackend]
        Reference to the OpticStudio backend.
    coordinates : Iterable[tuple[float, float]], optional
        An iterable of tuples representing the coordinates for the ray trace.
        If `field_type` is "angle", the coordinates should be the angles along the (X, Y) axes in degrees.
        If `field_type` is "object_height", the coordinates should be the object heights along the
        (X, Y) axes in mm. Defaults to `None`, which uses the fields defined in the backend.
    wavelengths : Iterable[float], optional
        An iterable of wavelengths to be used in the ray trace. Defaults to `None`, which uses the wavelengths
        defined in the backend.
    field_type : Literal["angle", "object_height"], optional
        The type of field to be used in the ray trace. Can be either "angle" or "object_height". Defaults to
        "angle".
    pupil : tuple[float, float], optional
        A tuple representing the pupil coordinates for the ray trace. Defaults to (0, 0).

    Returns
    -------
    DataFrame
        A pandas DataFrame containing the results of the ray trace analysis.
    """
    if abs(pupil[0]) > 1 or abs(pupil[1]) > 1:
        raise ValueError("Pupil coordinates must be between -1 and 1.")

    if coordinates is not None:
        backend.set_fields(coordinates, field_type=field_type)

    if wavelengths is not None:
        backend.set_wavelengths(wavelengths)

    real_ray_traces = []
    raytrace_results = []

    for wavelength_number, wavelength in backend.iter_wavelengths():
        for field_number, field in backend.iter_fields():
            raytrace_result = zp.analyses.raysandspots.SingleRayTrace(
                px=pupil[0],
                py=pupil[1],
                field=field_number,
                wavelength=wavelength_number,
                global_coordinates=True,
            ).run(backend.get_oss())
            real_ray_trace = raytrace_result.data.real_ray_trace_data

            if real_ray_trace is None:
                raise ValueError(
                    f"Failed to perform ray trace for field ({field.X}, {field.Y}) and wavelength {wavelength}."
                )

            real_ray_trace.insert(0, "Field", [(field.X, field.Y)] * len(real_ray_trace))
            real_ray_trace.insert(0, "Wavelength", wavelength)

            real_ray_traces.append(real_ray_trace)
            raytrace_results.append(raytrace_result.data)

    return _build_raytrace_result(real_ray_traces), raytrace_results



"""Optical surfaces for OpticStudio."""

from __future__ import annotations

from abc import ABC
from functools import singledispatch
from typing import TYPE_CHECKING, Generic, TypeVar, Union
from warnings import warn

import zospy as zp

from visisipy.models import BaseSurface
from visisipy.models import NoSurface as OpticStudioNoSurface
from visisipy.models.geometry import (
    NoSurface,
    StandardSurface,
    Stop,
    Surface,
    ZernikeCoefficients,
    ZernikeStandardPhaseSurface,
    ZernikeStandardSagSurface,
)
from visisipy.models.materials import MaterialModel

if TYPE_CHECKING:
    from zospy.api import _ZOSAPI
    from zospy.zpcore import OpticStudioSystem

__all__ = ("OpticStudioSurface", "OpticStudioZernikeStandardSagSurface", "make_surface")

PropertyType = TypeVar("PropertyType")


class OpticStudioSurfaceProperty(Generic[PropertyType]):
    """Descriptor for OpticStudio surface properties.

    This descriptor is used to access and modify properties of the OpticStudio surface.
    """

    def __init__(self, name: str) -> None:
        self.name = name

    def __get__(self, obj: OpticStudioSurface, objtype=None) -> PropertyType:
        if obj.surface is None:
            return None

        return getattr(obj.surface, self.name)

    def __set__(self, obj: OpticStudioSurface, value: PropertyType) -> None:
        if obj.surface is None:
            message = f"Cannot set attribute {self.name} of non-built surface."
            raise AttributeError(message)

        setattr(obj.surface, self.name, value)


class OpticStudioSurfaceDataProperty(Generic[PropertyType]):
    """Property descriptor for OpticStudio surface data properties.

    This descriptor is used to access and modify properties of the OpticStudio surface data.
    """

    def __init__(self, name: str) -> None:
        self.name = name

    def __get__(self, obj: OpticStudioSurface, objtype=None) -> PropertyType:
        if obj.surface is None:
            return None

        return getattr(obj.surface.SurfaceData, self.name)

    def __set__(self, obj: OpticStudioSurface, value: PropertyType) -> None:
        if obj.surface is None:
            message = f"Cannot set attribute {self.name} of non-built surface."
            raise AttributeError(message)

        setattr(obj.surface.SurfaceData, self.name, value)


class OpticStudioSurface(BaseSurface):
    """Sequential surface in OpticStudio."""

    _TYPE: str = "Standard"

    def __init__(
        self,
        comment: str,
        *,
        radius: float = float("inf"),
        thickness: float = 0.0,
        semi_diameter: float | None = None,
        conic: float = 0.0,
        material: MaterialModel | str | None = None,
        is_stop: bool | None = None,
    ):
        """Create a new OpticStudio surface.

        Parameters
        ----------
        comment : str
            Comment for the surface.
        radius : float
            Radius of curvature of the surface, in mm. Defaults to infinity (flat surface).
        thickness : float
            Thickness of the surface, in mm. Defaults to 0.0 mm.
        semi_diameter : float, optional
            Semi-diameter of the surface aperture, in mm. Defaults to `None`, which means the semi-diameter will be
            determined by OpticStudio.
        conic : float
            Conic constant of the surface. Defaults to 0.0 (spherical surface).
        material : MaterialModel | str, optional
            Material of the surface. This can be either a string representing the name of the material or a
            MaterialModel instance. Defaults to `None`, which means no material is set.
        is_stop : bool, optional
            If `True`, the surface is treated as a stop. Defaults to `None`, which means the stop status will be
            determined by OpticStudio. Note that setting the `IsStop` property in OpticStudio will always convert the
            surface to a stop, regardless of the value. When set to `False`, this parameter will be ignored.
        """
        self._comment = comment
        self._radius = radius
        self._thickness = thickness
        self._semi_diameter = semi_diameter
        self._conic = conic
        self._material = material
        self._is_stop = is_stop

        self._surface = None
        self._is_built = False

    comment: str = OpticStudioSurfaceProperty("Comment")
    radius: float = OpticStudioSurfaceProperty("Radius")
    thickness: float = OpticStudioSurfaceProperty("Thickness")
    semi_diameter: float = OpticStudioSurfaceProperty("SemiDiameter")
    conic: float = OpticStudioSurfaceProperty("Conic")
    is_stop: bool = OpticStudioSurfaceProperty("IsStop")

    def _get_material(self) -> MaterialModel | str | None:
        if not self._is_built:
            return None

        if self.surface.MaterialCell.GetSolveData().Type == zp.constants.Editors.SolveType.MaterialModel:
            material_model = self.surface.MaterialCell.GetSolveData()._S_MaterialModel  # noqa: SLF001

            return MaterialModel(
                refractive_index=material_model.IndexNd,
                abbe_number=material_model.AbbeVd,
                partial_dispersion=material_model.dPgF,
            )

        return self.surface.Material

    def _set_material(self, material: MaterialModel | str | None) -> None:
        if material is None:  # Do nothing if material is None
            return

        if isinstance(material, MaterialModel):
            zp.solvers.material_model(
                self.surface.MaterialCell,
                refractive_index=material.refractive_index,
                abbe_number=material.abbe_number,
                partial_dispersion=material.partial_dispersion,
            )
        elif isinstance(material, str):
            self.surface.Material = material
        else:
            raise TypeError("'material' must be MaterialModel or str.")

    @property
    def material(self) -> MaterialModel | str:
        """Material of the surface."""
        return self._get_material()

    @material.setter
    def material(self, material: MaterialModel | str) -> None:
        if self._is_built:
            self._set_material(material)

    @property
    def surface(self) -> _ZOSAPI.Editors.LDE.ILDERow | None:
        """OpticStudio surface object.

        This property only has a value if the surface has been built with `Surface.build`.
        """
        return self._surface

    def _set_surface_type(self):
        surface_type = zp.constants.process_constant(zp.constants.Editors.LDE.SurfaceType, self._TYPE)

        if self.surface is not None and self.surface.Type != surface_type:
            zp.functions.lde.surface_change_type(self.surface, self._TYPE)

    def build(self, oss: OpticStudioSystem, *, position: int, replace_existing: bool = False) -> int:
        """Create the surface in OpticStudio.

        Create the surface in the provided `OpticStudioSystem` `oss` at index `position`.
        By default, a new surface will be created. An existing surface will be overwritten if `replace_existing`
        is set to `True`.

        Parameters
        ----------
        oss : zospy.zpcore.OpticStudioSystem
            OpticStudio system in which the surface is created.
        position : int
            Index at which the surface is located, starting at 0 for the object surface.
        replace_existing : bool
            If `True`, replace an existing surface instead of inserting a new one. Defaults to `False`.

        Returns
        -------
        int
            The index of the created surface. Subsequent surfaces should be added after this index.
        """
        self._surface = oss.LDE.GetSurfaceAt(position) if replace_existing else oss.LDE.InsertNewSurfaceAt(position)

        self._set_surface_type()

        self.comment = self._comment
        self.radius = self._radius
        self.thickness = self._thickness
        self.conic = self._conic

        self._set_material(self._material)

        # Only set semi_diameter when explicitly specified
        if self._semi_diameter is not None:
            self.semi_diameter = self._semi_diameter

        # Only set IsStop when explicitly specified
        if self._is_stop is True:
            self.is_stop = self._is_stop
        elif self._is_stop is False:
            warn(
                "is_stop is set to False, but this is not supported in OpticStudio. Explicitly setting is_stop will "
                "always convert the surface to a stop. This setting has been ignored."
            )

        self._is_built = True

        return self.surface.SurfaceNumber

    def relink_surface(self, oss: OpticStudioSystem) -> bool:
        """Link an OpticStudio surface based on its comment.

        Searches for a surface in `oss` whose comment matches `self.comment`.
        This surface is assigned to `self.surface`. If no surface is found or multiple
        surfaces are found, `self.surface` is not updated.

        Parameters
        ----------
        oss : zospy.zpcore.OpticStudioSystem
            OpticStudio system in which the eye model is defined.

        Returns
        -------
        bool
            True if the operation succeeded, False if the system has not been built, or
            no / multiple surfaces have been found.
        """
        if self._is_built:
            surfaces = zp.functions.lde.find_surface_by_comment(oss.LDE, self._comment)

            if len(surfaces) == 1:
                self._surface = surfaces[0]

                return True

        return False


class BaseOpticStudioZernikeSurface(OpticStudioSurface, ABC):
    """Base class for Zernike surfaces in OpticStudio.

    This class provides methods and properties shared by all Zernike surfaces.
    """

    def __new__(cls, *args, **kwargs):  # noqa: ARG004, RUF100
        if cls is BaseOpticStudioZernikeSurface:
            raise TypeError("Only child classes of BaseOpticStudioZernikeSurface may be instantiated.")

        return super().__new__(cls)

    def __init__(
        self,
        comment: str,
        *,
        radius: float = float("inf"),
        thickness: float = 0.0,
        semi_diameter: float | None = None,
        conic: float = 0.0,
        material: MaterialModel | str | None = None,
        is_stop: bool | None = None,
        number_of_terms: int = 0,
        norm_radius: float = 100,
        zernike_coefficients: ZernikeCoefficients | None = None,
    ):
        super().__init__(
            comment=comment,
            radius=radius,
            thickness=thickness,
            semi_diameter=semi_diameter,
            conic=conic,
            material=material,
            is_stop=is_stop,
        )

        if zernike_coefficients is not None:
            if any(key > number_of_terms for key in zernike_coefficients):
                raise ValueError(f"Zernike coefficients must be smaller than the maximum term {number_of_terms}.")
            if any(key < 1 for key in zernike_coefficients):
                raise ValueError("Zernike coefficients must be larger than 0.")

        self._number_of_terms = number_of_terms
        self._norm_radius = norm_radius
        self._zernike_coefficients = zernike_coefficients or ZernikeCoefficients()

    number_of_terms: int = OpticStudioSurfaceDataProperty("NumberOfTerms")
    norm_radius: float = OpticStudioSurfaceDataProperty("NormRadius")

    def _validate_coefficient(self, n: int):
        if n < 1:
            raise ValueError("Zernike coefficient must be larger than 0.")
        if n > self.number_of_terms:
            raise ValueError(f"Zernike coefficient must be smaller than the maximum term {self.number_of_terms}.")

    def get_zernike_coefficient(self, n: int) -> float:
        """Get the value of the nth Zernike coefficient.

        Parameters
        ----------
        n : int
            The Zernike coefficient to retrieve.

        Returns
        -------
        float
            The value of the Zernike coefficient.

        Raises
        ------
        ValueError
            If `n` is less than 0 or larger than the maximum term.
        """
        self._validate_coefficient(n)

        return self.surface.SurfaceData.GetNthZernikeCoefficient(n)

    def set_zernike_coefficient(self, n: int, value: float) -> None:
        """Set the value of the nth Zernike coefficient.

        Parameters
        ----------
        n : int
            The Zernike coefficient to set.
        value : float
            The value of the Zernike coefficient.

        Raises
        ------
        ValueError
            If `n` is less than 0 or larger than the maximum term.
        """
        self._validate_coefficient(n)

        self.surface.SurfaceData.SetNthZernikeCoefficient(n, value)

    def build(self, oss: OpticStudioSystem, *, position: int, replace_existing: bool = False) -> int:
        """Create the surface in OpticStudio.

        Create the surface in the provided `OpticStudioSystem` `oss` at index `position`.
        By default, a new surface will be created. An existing surface will be overwritten if `replace_existing`
        is set to `True`.

        Parameters
        ----------
        oss : zospy.zpcore.OpticStudioSystem
            OpticStudio system in which the surface is created.
        position : int
            Index at which the surface is located, starting at 0 for the object surface.
        replace_existing : bool
            If `True`, replace an existing surface instead of inserting a new one. Defaults to `False`.

        Returns
        -------
        int
            The index of the created surface. Subsequent surfaces should be added after this index.
        """
        index = super().build(oss, position=position, replace_existing=replace_existing)

        self.number_of_terms = self._number_of_terms
        self.norm_radius = self._norm_radius

        for n, value in self._zernike_coefficients.items():
            self.set_zernike_coefficient(n, value)

        return index


class OpticStudioZernikeStandardSagSurface(BaseOpticStudioZernikeSurface):
    """Zernike Standard Sag surface in OpticStudio."""

    _TYPE = "ZernikeStandardSag"

    def __init__(
        self,
        comment: str,
        *,
        radius: float = float("inf"),
        thickness: float = 0.0,
        semi_diameter: float | None = None,
        conic: float = 0.0,
        material: MaterialModel | str | None = None,
        is_stop: bool | None = None,
        extrapolate: int = 0,
        zernike_decenter_x: float = 0.0,
        zernike_decenter_y: float = 0.0,
        number_of_terms: int = 0,
        norm_radius: float = 100,
        zernike_coefficients: ZernikeCoefficients | None = None,
    ):
        super().__init__(
            comment=comment,
            radius=radius,
            thickness=thickness,
            semi_diameter=semi_diameter,
            conic=conic,
            material=material,
            is_stop=is_stop,
            number_of_terms=number_of_terms,
            norm_radius=norm_radius,
            zernike_coefficients=zernike_coefficients,
        )

        self._extrapolate = extrapolate
        self._zernike_decenter_x = zernike_decenter_x
        self._zernike_decenter_y = zernike_decenter_y

    extrapolate: int = OpticStudioSurfaceDataProperty("Extrapolate")
    zernike_decenter_x: float = OpticStudioSurfaceDataProperty("ZernikeDecenter_X")
    zernike_decenter_y: float = OpticStudioSurfaceDataProperty("ZernikeDecenter_Y")

    def build(self, oss: OpticStudioSystem, *, position: int, replace_existing: bool = False) -> int:
        """Create the surface in OpticStudio.

        Create the surface in the provided `OpticStudioSystem` `oss` at index `position`.
        By default, a new surface will be created. An existing surface will be overwritten if `replace_existing`
        is set to `True`.

        Parameters
        ----------
        oss : zospy.zpcore.OpticStudioSystem
            OpticStudio system in which the surface is created.
        position : int
            Index at which the surface is located, starting at 0 for the object surface.
        replace_existing : bool
            If `True`, replace an existing surface instead of inserting a new one. Defaults to `False`.

        Returns
        -------
        int
            The index of the created surface. Subsequent surfaces should be added after this index.
        """
        index = super().build(oss, position=position, replace_existing=replace_existing)

        self.extrapolate = self._extrapolate
        self.zernike_decenter_x = self._zernike_decenter_x
        self.zernike_decenter_y = self._zernike_decenter_y

        return index


class OpticStudioZernikeStandardPhaseSurface(BaseOpticStudioZernikeSurface):
    """Zernike Standard Phase surface in OpticStudio."""

    _TYPE = "ZernikeStandardPhase"

    def __init__(
        self,
        comment: str,
        *,
        radius: float = float("inf"),
        thickness: float = 0.0,
        semi_diameter: float | None = None,
        conic: float = 0.0,
        material: MaterialModel | str | None = None,
        is_stop: bool | None = None,
        extrapolate: int = 0,
        diffract_order: float = 0.0,
        number_of_terms: int = 0,
        norm_radius: float = 100,
        zernike_coefficients: ZernikeCoefficients | None = None,
    ):
        super().__init__(
            comment=comment,
            radius=radius,
            thickness=thickness,
            semi_diameter=semi_diameter,
            conic=conic,
            material=material,
            is_stop=is_stop,
            number_of_terms=number_of_terms,
            norm_radius=norm_radius,
            zernike_coefficients=zernike_coefficients,
        )

        self._extrapolate = extrapolate
        self._diffract_order = diffract_order

    extrapolate: int = OpticStudioSurfaceDataProperty("Extrapolate")
    diffract_order: float = OpticStudioSurfaceDataProperty("DiffractOrder")

    def build(self, oss: OpticStudioSystem, *, position: int, replace_existing: bool = False) -> int:
        """Create the surface in OpticStudio.

        Create the surface in the provided `OpticStudioSystem` `oss` at index `position`.
        By default, a new surface will be created. An existing surface will be overwritten if `replace_existing`
        is set to `True`.

        Parameters
        ----------
        oss : zospy.zpcore.OpticStudioSystem
            OpticStudio system in which the surface is created.
        position : int
            Index at which the surface is located, starting at 0 for the object surface.
        replace_existing : bool
            If `True`, replace an existing surface instead of inserting a new one. Defaults to `False`.

        Returns
        -------
        int
            The index of the created surface. Subsequent surfaces should be added after this index.
        """
        index = super().build(oss, position=position, replace_existing=replace_existing)

        self.extrapolate = self._extrapolate
        self.diffract_order = self._diffract_order

        return index


@singledispatch
def make_surface(surface: Surface, material: str | MaterialModel, comment: str = "") -> OpticStudioSurface:
    """Create an `OpticStudioSurface` instance from a given `Surface` instance.

    Parameters
    ----------
    surface : Surface
        The Surface instance from which to create the OpticStudioSurface instance.
    material : str | MaterialModel
        The material of the surface. This can be either a string representing the name
        of the material or a MaterialModel instance.
    comment : str, optional
        A comment to be associated with the surface. This is an empty string by default.

    Returns
    -------
    OpticStudioSurface
        The created OpticStudioSurface instance.
    """
    return OpticStudioSurface(comment=comment, thickness=surface.thickness, material=material)


@make_surface.register
def _make_surface(
    surface: StandardSurface,
    material: Union[str, MaterialModel],  # noqa: UP007
    comment: str = "",
) -> OpticStudioSurface:
    return OpticStudioSurface(
        comment=comment,
        radius=surface.radius,
        thickness=surface.thickness,
        semi_diameter=surface.semi_diameter,
        conic=surface.asphericity,
        material=material,
    )


@make_surface.register
def _make_surface(
    surface: Stop,
    material: Union[str, MaterialModel] = "",  # noqa: UP007
    comment: str = "",
) -> OpticStudioSurface:
    return OpticStudioSurface(
        comment=comment,
        thickness=surface.thickness,
        material=material,
        semi_diameter=surface.semi_diameter,
        is_stop=True,
    )


@make_surface.register
def _make_surface(
    surface: ZernikeStandardSagSurface,
    material: Union[str, MaterialModel] = "",  # noqa: UP007
    comment: str = "",
) -> OpticStudioZernikeStandardSagSurface:
    return OpticStudioZernikeStandardSagSurface(
        comment=comment,
        radius=surface.radius,
        thickness=surface.thickness,
        semi_diameter=surface.semi_diameter,
        conic=surface.asphericity,
        material=material,
        zernike_coefficients=surface.zernike_coefficients,
        extrapolate=1 if surface.extrapolate else 0,
        zernike_decenter_x=surface.zernike_decenter_x,
        zernike_decenter_y=surface.zernike_decenter_y,
        number_of_terms=surface.maximum_term,
        norm_radius=surface.norm_radius,
    )


@make_surface.register
def _make_surface(
    surface: ZernikeStandardPhaseSurface,
    material: Union[str, MaterialModel] = "",  # noqa: UP007
    comment: str = "",
) -> OpticStudioZernikeStandardPhaseSurface:
    return OpticStudioZernikeStandardPhaseSurface(
        comment=comment,
        radius=surface.radius,
        thickness=surface.thickness,
        semi_diameter=surface.semi_diameter,
        conic=surface.asphericity,
        material=material,
        zernike_coefficients=surface.zernike_coefficients,
        extrapolate=1 if surface.extrapolate else 0,
        diffract_order=surface.diffraction_order,
        number_of_terms=surface.maximum_term,
        norm_radius=surface.norm_radius,
    )


@make_surface.register
def _make_surface(
    surface: NoSurface,  # noqa: ARG001
    material: None = None,  # noqa: ARG001
    comment: str = "",  # noqa: ARG001
) -> OpticStudioNoSurface:
    return OpticStudioNoSurface()



from __future__ import annotations

import pytest
from optiland.optic import Optic


@pytest.fixture
def optic() -> Optic:
    """Fixture to create an Optic instance for testing."""
    system = Optic()

    system.set_field_type("angle")
    system.add_field(0, 0)
    system.add_wavelength(0.543)
    system.set_aperture("EPD", 1.0)

    return system



from __future__ import annotations

from types import SimpleNamespace
from typing import TYPE_CHECKING

import pytest

from visisipy import EyeModel
from visisipy.analysis import base

if TYPE_CHECKING:
    from visisipy.backend import BaseBackend


class MockBackend:
    def __init__(self):
        self.model = None

    def build_model(self, model):
        self.model = SimpleNamespace(eye_model=model)


@pytest.fixture
def mock_backend(monkeypatch):
    """Mock the backend for testing."""

    backend = MockBackend()

    monkeypatch.setattr(base, "get_backend", lambda: backend)

    return backend


@pytest.fixture
def example_analysis():
    """Example analysis function to test the analysis decorator."""

    def example_analysis(model: EyeModel | None, x: int, *, return_raw_result: bool, backend: type[BaseBackend]):
        return x, x

    return example_analysis


class TestAnalysisDecorator:
    def test_decorator(self, mock_backend, example_analysis):
        decorated_analysis = base.analysis(example_analysis)

        assert decorated_analysis(EyeModel(), 1) == 1
        assert decorated_analysis(EyeModel(), 1, return_raw_result=True) == (1, 1)

    def test_build_model(self, mock_backend, example_analysis):
        model = EyeModel()

        decorated_analysis = base.analysis(example_analysis)
        decorated_analysis(model, 1)

        assert mock_backend.model is not None
        assert mock_backend.model.eye_model == model

    def test_get_backend(self, mock_backend):
        def example_analysis(
            model: EyeModel | None, x: int, *, return_raw_result: bool, backend: type[BaseBackend]
        ) -> tuple[type[BaseBackend], int]:
            return backend, x

        decorated_analysis = base.analysis(example_analysis)

        assert decorated_analysis(EyeModel(), 1) == mock_backend

    def test_pass_backend(self):
        def example_analysis(
            model: EyeModel | None, x: int, *, return_raw_result: bool, backend: type[BaseBackend]
        ) -> tuple[type[BaseBackend], int]:
            return backend, x

        decorated_analysis = base.analysis(example_analysis)
        backend = MockBackend()

        assert decorated_analysis(EyeModel(), 1, backend=backend) == backend

    def test_no_model_raises_valueerror(self):
        def example_analysis(x, *, return_raw_result: bool = False, backend: type[BaseBackend]):
            return x, x

        with pytest.raises(
            ValueError,
            match="The first parameter of an analysis function must be 'model'",
        ):
            base.analysis(example_analysis)

    def test_invalid_model_annotation_raises_valueerror(self):
        def example_analysis(model: str, x, *, return_raw_result: bool = False, backend: type[BaseBackend]):
            return x, x

        with pytest.raises(
            ValueError,
            match="The 'model' parameter of an analysis function must have type 'EyeModel | None', got 'str'",
        ):
            base.analysis(example_analysis)

    def test_no_return_raw_result_raises_valueerror(self):
        def example_analysis(model: EyeModel | None, x, *, backend: type[BaseBackend]):
            return x, x

        with pytest.raises(
            ValueError,
            match="The analysis function must have a keyword-only 'return_raw_result' parameter of type 'bool'",
        ):
            base.analysis(example_analysis)

    def test_no_keyword_only_return_raw_result_raises_valueerror(self):
        def example_analysis(
            model: EyeModel | None,
            x,
            return_raw_result: bool,
            *,
            backend: type[BaseBackend],
        ):
            return x, x

        with pytest.raises(
            ValueError,
            match="The 'return_raw_result' parameter of an analysis function must be keyword-only",
        ):
            base.analysis(example_analysis)

    def test_invalid_return_raw_result_annotation_raises_valueerror(self):
        def example_analysis(model: EyeModel | None, x, *, return_raw_result: str, backend: type[BaseBackend]):
            return x, x

        with pytest.raises(
            ValueError,
            match="The 'return_raw_result' parameter of an analysis function must have type 'bool', got 'str'",
        ):
            base.analysis(example_analysis)

    def test_no_backend_raises_valueerror(self):
        def example_analysis(model: EyeModel | None, x, *, return_raw_result: bool):
            return x, x

        with pytest.raises(
            ValueError,
            match=r"The analysis function must have a keyword-only 'backend' parameter of type 'type\[BaseBackend\]'",
        ):
            base.analysis(example_analysis)

    def test_no_keyword_only_backend_raises_valueerror(self):
        def example_analysis(model: EyeModel | None, x, backend: type[BaseBackend], *, return_raw_result: bool):
            return x, x

        with pytest.raises(
            ValueError,
            match=r"The 'backend' parameter of an analysis function must be keyword-only",
        ):
            base.analysis(example_analysis)

    def test_invalid_backend_annotation_raises_valueerror(self):
        def example_analysis(model: EyeModel | None, x, *, return_raw_result: bool, backend: str):
            return x, x

        with pytest.raises(
            ValueError,
            match=r"The 'backend' parameter of an analysis function must have type 'type\[BaseBackend\]', "
            r"got 'str'",
        ):
            base.analysis(example_analysis)



"""Utilities for handling wavefront data."""

from __future__ import annotations

from collections import defaultdict

__all__ = ("ZernikeCoefficients",)


def _is_int_tuple(x: tuple[int, int]) -> bool:
    return isinstance(x, tuple) and len(x) == 2 and all(isinstance(i, int) for i in x)  # noqa: PLR2004


def _validate_coefficient(key: int | tuple[int, int]) -> int:
    if isinstance(key, int):
        if key < 1:
            raise ValueError("The coefficient must be larger than 0.")

        return key

    if _is_int_tuple(key):
        _validate_nm(*key)
        return ZernikeCoefficients.to_noll(*key)

    raise TypeError("The coefficient must be an integer or a tuple of two integers.")


def _validate_nm(n: int, m: int) -> None:
    if n < 0:
        raise ValueError("n must be greater than or equal to 0.")
    if abs(m) > n:
        raise ValueError("m must be less than or equal to n.")
    if (n - abs(m)) % 2 != 0:
        raise ValueError("n and m must have the same parity.")


class ZernikeCoefficients(defaultdict):
    """Zernike coefficients.

    Convenience class for handling Zernike coefficients as a dictionary. If a term is not present, 0 is returned.
    Upon initialization and setting items, the keys are validated to be non-negative integers.

    Raises
    ------
    TypeError
        If the key is not an integer.
    ValueError
        If the key is smaller than 1.
    """

    def __init__(self, terms: dict[int, float] | None = None):
        if terms is not None:
            if not all(isinstance(key, int) for key in terms):
                raise TypeError("All keys must be integers.")
            if any(key < 1 for key in terms):
                raise ValueError("The Zernike coefficients must be larger than 0.")

        super().__init__(lambda: 0, terms or {})

    @staticmethod
    def to_noll(n: int, m: int):
        """Convert Zernike indices to Noll indices.

        Parameters
        ----------
        n : int
            The radial order of the Zernike polynomial.
        m : int
            The azimuthal order of the Zernike polynomial.

        Returns
        -------
        int
            The Noll index corresponding to the given Zernike indices.
        """
        # Validate n and m
        _validate_nm(n, m)

        i = n % 4 in {0, 1}
        k = 0 if (m > 0 and i) or (m < 0 and not i) else 1

        return ((n + 1) * n) // 2 + abs(m) + k

    def __setitem__(self, key: int | tuple[int, int], value: float) -> None:
        key = _validate_coefficient(key)

        super().__setitem__(key, value)

    def __getitem__(self, key: int | tuple[int, int]) -> float:
        key = _validate_coefficient(key)

        return super().__getitem__(key)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({dict(self)})"


def min_max_noll_index(min_order: int, max_order: int) -> tuple[int, int]:
    """Get the minimum and maximum Noll indices for a given range of Zernike orders.

    Parameters
    ----------
    min_order : int
        The minimum Zernike order.
    max_order : int
        The maximum Zernike order.

    Returns
    -------
    tuple[int, int]
        The minimum and maximum Noll indices corresponding to the given Zernike orders.
    """
    min_index = ((min_order + 1) * min_order) // 2 + 1
    max_index = ((max_order + 1) * max_order) // 2 + max_order + 1

    return min_index, max_index



"""Zernike coefficients analysis for Optiland."""

from __future__ import annotations

from typing import TYPE_CHECKING

from optiland.wavefront import ZernikeOPD

from visisipy.types import SampleSize
from visisipy.wavefront import ZernikeCoefficients

if TYPE_CHECKING:
    from visisipy.backend import FieldCoordinate, FieldType
    from visisipy.optiland.backend import OptilandBackend


def _build_zernike_coefficients(
    zernike_opd: ZernikeOPD,
) -> ZernikeCoefficients:
    return ZernikeCoefficients(dict(enumerate(zernike_opd.coeffs, start=1)))


def zernike_standard_coefficients(
    backend: type[OptilandBackend],
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 64,
    maximum_term: int = 45,
) -> tuple[ZernikeCoefficients, ZernikeOPD]:
    """Calculate the Zernike standard coefficients at the retina surface.

    Parameters
    ----------
    backend : type[OptilandBackend]
        Reference to the Optiland backend.
    field_coordinate : tuple[float, float] | None, optional
        The field coordinate for the Zernike calculation. When `None`, the first field in OpticStudio is used.
        Defaults to `None`.
    wavelength : float | None, optional
        The wavelength for the Zernike calculation. When `None`, the first wavelength in OpticStudio is used.
        Defaults to `None`.
    field_type : Literal["angle", "object_height"], optional
        The type of field to be used when setting the field coordinate. This parameter is only used when
        `field_coordinate` is specified. Defaults to "angle".
    sampling : SampleSize | str | int, optional
        The sampling for the Zernike calculation. Defaults to 512.
    maximum_term : int, optional
        The maximum term for the Zernike calculation. Defaults to 45.

    Returns
    -------
    AttrDict
        ZOSPy Zernike standard coefficients analysis output.
    """
    if not isinstance(sampling, SampleSize):
        sampling = SampleSize(sampling)

    if field_coordinate is not None:
        backend.set_fields([field_coordinate], field_type=field_type)

    if wavelength is None:
        wavelength = backend.get_wavelengths()[0]
    else:
        backend.set_wavelengths([wavelength])

    normalized_field = backend.get_optic().fields.get_field_coords()[0]

    zernike_opd = ZernikeOPD(
        backend.optic,
        field=normalized_field,
        wavelength=wavelength,
        num_rings=int(sampling),
        num_terms=maximum_term,
        zernike_type="noll",
    )

    return _build_zernike_coefficients(zernike_opd), zernike_opd



"""Optical eye models.

This module provides the base classes for eye models and surfaces.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

import visisipy.backend as _backend
from visisipy.models.geometry import EyeGeometry, NavarroGeometry, create_geometry
from visisipy.models.materials import (
    EyeMaterials,
    NavarroMaterials,
    NavarroMaterials458,
    NavarroMaterials543,
    NavarroMaterials589,
    NavarroMaterials633,
)

if TYPE_CHECKING:
    from os import PathLike

__all__ = (
    "BaseEye",
    "BaseSurface",
    "EyeGeometry",
    "EyeMaterials",
    "EyeModel",
    "NavarroGeometry",
    "NavarroMaterials",
    "NavarroMaterials458",
    "NavarroMaterials543",
    "NavarroMaterials589",
    "NavarroMaterials633",
    "create_geometry",
)


@dataclass
class EyeModel:
    """Optical model of the eye.

    Visisipy's eye models consist of two parts: the geometry and the material model. The geometry defines the shape of
    the eye, while the material model defines the optical properties of the materials in the eye. By default, this model
    uses the geometry and materials of the Navarro wide-field eye model [1]_. The default material model has been
    fitted to the refractive indices reported by Escudero-Sanz and Navarro [1]_. This model will work with all visible
    wavelengths, but could deviate slightly from the values provided in the literature for the specified wavelengths.

    Attributes
    ----------
    geometry : EyeGeometry
        Geometry of the eye. Defaults to the geometry of the Navarro eye model.
    materials : EyeMaterials
        Properties of the materials of the eye. Defaults to the materials of the Navarro eye model.

    See Also
    --------
    NavarroGeometry : Geometry of the Navarro eye model.
    NavarroMaterials : Materials of the Navarro eye model.

    References
    ----------
    .. [1] Escudero-Sanz, I., & Navarro, R. (1999).
       Off-axis aberrations of a wide-angle schematic eye model.
       JOSA A, 16(8), 1881-1891. https://doi.org/10.1364/JOSAA.16.001881
    """

    geometry: EyeGeometry = field(default_factory=NavarroGeometry)
    materials: EyeMaterials = field(default_factory=NavarroMaterials)
    _built: BaseEye | None = field(default=None, init=False, repr=False)

    def build(
        self,
        *,
        start_from_index: int = 0,
        replace_existing: bool = False,
        object_distance: float = float("inf"),
        **kwargs,
    ) -> BaseEye:
        """Build the eye model in the backend.

        If no backend has yet been initialized, the default backend is used. The eye model is built in the backend
        starting from the surface at index `start_from_index`. The cornea front surface will be located at
        `start_from_index + 1`. If `replace_existing` is set to `True`, the existing model in the backend will be
        overwritten. The `object_distance` parameter is used to set the distance between the cornea front and the
        surface preceding the eye model.

        Parameters
        ----------
        start_from_index : int
            Index of the surface after which the eye model will be built.
        replace_existing : bool
            If `True`, the existing model in the backend will be overwritten.
        object_distance : float
            Distance between the cornea front and the surface preceding the eye model.
        """
        backend = _backend.get_backend()
        self._built = backend.build_model(
            self,
            start_from_index=start_from_index,
            replace_existing=replace_existing,
            object_distance=object_distance,
            **kwargs,
        )

        return self._built

    @staticmethod
    def clear() -> None:
        """Clear the model in the backend."""
        backend = _backend.get_backend()
        backend.clear_model()

    @staticmethod
    def save(filename: str | PathLike | None = None) -> None:
        """Save the model definition to a file.

        Saving is done by the backend. As a consequence, the file format depends on the backend used.

        Parameters
        ----------
        filename : str | PathLike | None
            Name of the file to save the model. If `None`, the model will be saved in the default location.
            The file format depends on the backend used.
        """
        backend = _backend.get_backend()
        backend.save_model(filename)


class BaseSurface(ABC):
    """Abstract class that must be implemented by backend-specific surface classes."""

    @abstractmethod
    def __init__(
        self,
        comment: str,
        radius: float = float("inf"),
        thickness: float = 0.0,
        semi_diameter: float | None = None,
        conic: float = 0.0,
        material: Any | None = None,
        *,
        is_stop: bool | None = None,
    ): ...

    @property
    @abstractmethod
    def surface(self) -> Any: ...

    @abstractmethod
    def build(self, *args, position: int, replace_existing: bool = False) -> int:
        """Build the surface in the backend."""
        ...


class NoSurface(BaseSurface):
    """Dummy surface class for when no surface is needed.

    This is a generic implementation that works with all backends, because it does not modify the optical system.
    """

    def __init__(self, *args, **kwargs):
        pass

    @property
    def surface(self) -> None:
        return None

    def build(self, *args, position: int, **kwargs) -> int:  # noqa: ARG002, PLR6301
        return position - 1


class BaseEye(ABC):
    """Abstract class that must be implemented by backend-specific eye model classes."""

    @abstractmethod
    def __init__(self, model: EyeModel): ...

    @abstractmethod
    def build(
        self,
        *args,
        start_from_index: int = 0,
        replace_existing: bool = False,
        object_distance: float = float("inf"),
        **kwargs,
    ):
        """Build the eye model in the backend.

        If no backend has yet been initialized, the default backend is used. The eye model is built in the backend
        starting from the surface at index `start_from_index`. The cornea front surface will be located at
        `start_from_index + 1`. If `replace_existing` is set to `True`, the existing model in the backend will be
        overwritten. The `object_distance` parameter is used to set the distance between the cornea front and the
        surface preceding the eye model.

        Parameters
        ----------
        start_from_index : int
            Index of the surface after which the eye model will be built.
        replace_existing : bool
            If `True`, the existing model in the backend will be overwritten.
        object_distance : float
            Distance between the cornea front and the surface preceding the eye model.
        """

    @property
    @abstractmethod
    def eye_model(self) -> EyeModel: ...

    @property
    def surfaces(self) -> dict[str, BaseSurface]:
        """Dictionary with surface names as keys and surfaces as values."""
        return {k.lstrip("_"): v for k, v in self.__dict__.items() if isinstance(v, BaseSurface)}

    def update_surfaces(self, attribute: str, value: Any, surface_names: list[str] | None = None) -> None:
        """Batch update all surfaces.

        Set `attribute` to `value` for multiple surfaces. If `surfaces` is not specified, all surfaces of the eye
        model are updated.

        Parameters
        ----------
        attribute : str
            Name of the attribute to update
        value : Any
            New value of the surface attribute
        surfaces : list[str]
            List of surfaces to be updated. If not specified, all surfaces are updated.

        Returns
        -------
        """
        surfaces = [self.surfaces[s] for s in surface_names] if surface_names is not None else self.surfaces.values()

        for s in surfaces:
            setattr(s, attribute, value)



from __future__ import annotations

from contextlib import nullcontext as does_not_raise

import pytest

from tests.helpers import build_args
from visisipy import EyeModel

pytestmark = [pytest.mark.needs_opticstudio]


class TestFFTPSFAnalysis:
    @pytest.mark.parametrize(
        "field_coordinate,wavelength,field_type,sampling",
        [
            (None, None, None, None),
            (None, None, "object_height", 64),
            ((0, 0), 0.550, "angle", 128),
            ((10, 5), 0.400, "object_height", 256),
            ((5.5, 5.5), 0.550, "angle", 512),
        ],
    )
    def test_fft_psf(
        self,
        opticstudio_backend,
        opticstudio_analysis,
        field_coordinate,
        wavelength,
        field_type,
        sampling,
    ):
        opticstudio_backend.build_model(
            EyeModel(), object_distance=10 if field_type == "object_height" else float("inf")
        )

        args = build_args(
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            field_type=field_type,
            sampling=sampling,
            non_null_defaults={"field_type", "sampling"},
        )

        assert opticstudio_analysis.fft_psf(**args)


class TestHuygensPSFAnalysis:
    @pytest.mark.parametrize(
        "field_coordinate,wavelength,field_type,pupil_sampling,image_sampling",
        [
            (None, None, None, None, None),
            (None, None, "object_height", 32, 32),
            ((0, 0), 0.550, "angle", 64, 128),
            ((10, 5), 0.400, "object_height", 256, 128),
            ((5.5, 5.5), 0.550, "angle", 64, 64),
        ],
    )
    def test_fft_psf(
        self,
        opticstudio_backend,
        opticstudio_analysis,
        field_coordinate,
        wavelength,
        field_type,
        pupil_sampling,
        image_sampling,
    ):
        opticstudio_backend.build_model(
            EyeModel(), object_distance=10 if field_type == "object_height" else float("inf")
        )

        args = build_args(
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            field_type=field_type,
            pupil_sampling=pupil_sampling,
            image_sampling=image_sampling,
            non_null_defaults={"field_type", "pupil_sampling", "image_sampling"},
        )

        assert opticstudio_analysis.huygens_psf(**args)


class TestStrehlRatioAnalysis:
    @pytest.mark.parametrize(
        "field_coordinate,wavelength,field_type,sampling",
        [
            (None, None, None, None),
            (None, None, "object_height", 64),
            ((0, 0), 0.550, "angle", 128),
            ((10, 5), 0.400, "object_height", 256),
            ((5.5, 5.5), 0.550, "angle", 64),
        ],
    )
    def test_strehl_ratio(
        self,
        opticstudio_backend,
        opticstudio_analysis,
        field_coordinate,
        wavelength,
        field_type,
        sampling,
    ):
        opticstudio_backend.build_model(
            EyeModel(), object_distance=10 if field_type == "object_height" else float("inf")
        )

        args = build_args(
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            field_type=field_type,
            sampling=sampling,
            non_null_defaults={"field_type", "sampling"},
        )

        assert opticstudio_analysis.strehl_ratio(**args, psf_type="huygens")

    @pytest.mark.parametrize(
        "psf_type,expectation",
        [
            (
                "fft",
                pytest.raises(
                    NotImplementedError,
                    match="OpticStudio does not support obtaining the Strehl ratio from the FFT PSF",
                ),
            ),
            ("huygens", does_not_raise()),
            (
                "invalid",
                pytest.raises(
                    NotImplementedError, match="PSF type 'invalid' is not implemented. Only 'huygens' is supported."
                ),
            ),
        ],
    )
    def test_psf_type(self, opticstudio_backend, opticstudio_analysis, psf_type, expectation):
        opticstudio_backend.build_model(EyeModel(), object_distance=float("inf"))

        with expectation:
            opticstudio_analysis.strehl_ratio(psf_type=psf_type)



from __future__ import annotations

from visisipy.models.materials import EyeMaterials, MaterialModel, NavarroMaterials


class TestMaterialModel:
    def test_initialization(self):
        material = MaterialModel(1.5, 1, 0.06)
        assert material.refractive_index == 1.5
        assert material.abbe_number == 1
        assert material.partial_dispersion == 0.06


class TestEyeMaterials:
    def test_initialization(self):
        cornea = MaterialModel(1.5, 50, 0.06)
        aqueous = MaterialModel(1.4, 45, 0.05)
        lens = MaterialModel(1.6, 55, 0.07)
        vitreous = MaterialModel(1.3, 40, 0.04)

        eye_materials = EyeMaterials(cornea, aqueous, lens, vitreous)

        assert eye_materials.cornea == cornea
        assert eye_materials.aqueous == aqueous
        assert eye_materials.lens == lens
        assert eye_materials.vitreous == vitreous


class TestNavarroMaterials:
    def test_default_initialization(self):
        navarro_materials = NavarroMaterials()
        assert navarro_materials.cornea.refractive_index == 1.3760893927086157
        assert navarro_materials.aqueous.refractive_index == 1.337454140357322
        assert navarro_materials.lens.refractive_index == 1.4201120875715691
        assert navarro_materials.vitreous.refractive_index == 1.3360870349145728
